! Module: omdetect_mod
!
! Authors:  Tim Harries - MSSL (orig) 
!           almost completely rewritten by J. Mittaz MSSL
!          
! Description: Uses a simple algorithm to detect sources in an OM OSW image
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!
! History
!
!    Created: Jan 1998 (TJH)
! 
!        July 1998: completely revised algorithm employed
!
!
!    27 April 1998: Removed diagnostic output
!                 : changed copy from rimage to image
!                 : removed datestamp call (TJH)
!
!    07 May   1998: Correct code for count rate bug (TJH)
!
!    Nov 2000 : Background by smoothing + Poisson noise for detection signif
!    Jan 2000 : Now uses mulitple scales + extractor type source 
!               parameterisation including deblending
!    24th August 2000 - AMK
!    Code added to read in the quality array.
!    The task will have to be modified eventually so that the quality array is processed.
!    We have seen images that have been generated by OMFLATFIELD that contain pixel values 
!    that contain values that have been set to the BLANK value. This occurrs due to the Flatfield
!    being zero for certain pixels. the quality array has been flagged to the value of 16 for these
!    cases. The image has been modified to contain the value zero as a temporary fix.
!    2nd. October 2000 - AMK
!    Reinstated the call to CAL_omGetPlateScale
!    3rd. Octeober 2000 - AMK
!    Replace the call getFIlterID in mssllib by the OAL call filterIdToInt
!    10th October 2000 - AMK Added White to list of filters - this can occur in the Summary file now
!    11th December (CB) - Background counts for each source added to the source list.
!    1st May 2001 (CB) - All pgplot drawing code removed. Limiting plate count rate calculated using 
!                        the photometry aperture radius. Code tidied up.
!*********************************************************************************************************** 
module omdetect_mod

    use dal
    use cal
    use param  
    use sort_mod
    use sasdal
    use omdetectnumerics_mod
    use omdetectsourceparams_mod
    use get_free_io_unit_mod
    use getFilterID_mod
    use errorHandling

    implicit none 
 
  private
  public  omdetect

    type srcpixel                             ! Pixel list node.
        integer(kind=int32) x                  ! X position.
        integer(kind=int32) y                  ! Y position.
        type(srcpixel), pointer :: next        ! Next pixel in source.
        type(srcpixel), pointer :: previous    ! Previous one (mem removal)
    end type srcpixel
  
    type src                                  ! Source list node.
        integer(kind=int32) sourceNum          ! Source number.
        integer(kind=int32) numPixels          ! Number of pixels.
        type(srcpixel), pointer :: pixelList   ! Pixel list.
        type(src), pointer :: next             ! Next source in list.
        type(src), pointer :: previous         ! previous one (memory removal)
    end type src


!
!---------------------------------------------------------------------
!
  
contains 
  
!---------------------------------------------------------------------
subroutine omdetect
!---------------------------------------------------------------------
    !*******************
    ! Filenames
    !*******************
    character(len=1024)  :: set                     ! filename of input image
    character(len=1024)  :: backgroundImage         ! Output background image
    character(len=1024)  :: levelImageName
    character(len=1024)  :: outset                  ! filename of osw source list
    character(len=1024)  :: flatField               ! filename of osw flat field
    character(len=1024)  :: mod8noise               ! filename of mod8noise
    character(len=1024)  :: regionfile              ! filename of mod8noise

    !****************
    ! DAL variables
    !****************
    type(DataSetT)       :: oswDataSet, listDataSet ! their associated handles
    type(DataSetT)       :: backgroundSet
    type(DataSetT)       :: levelImageSet 
    type(ArrayT)         :: oswArray                ! the images
    type(TableT)         :: listTable               ! source list handle
    type(ColumnT)        :: pcolumn                 ! pointer to a column
    type(ArrayT)         :: backgroundArray         ! array for background
    type(ArrayT)         :: levelImageArray         ! array for background

    ! input Poisson stats or not

    integer              :: boxScale                ! sliding box scale
    integer              :: maxScale                ! Maximum scale to search
    logical(kind=bool)   :: poisson

    ! downhill simplex variables

    ! variables of types in detectsource_mod
    !***********************
    ! binning variables
    !***********************
    integer                      :: binx, biny
    logical                      :: twixel

    character(len=6)            :: omFilter   ! the OM filter

    !**************************************
    ! variables for the output sourcelist
    !**************************************
    integer :: size = 50
   
    real, dimension(:), allocatable :: xsource             ! x source ps (pix)
    real, dimension(:), allocatable :: ysource             ! y source pos (px)
    real, dimension(:), allocatable :: sourcesignif        ! source counts
    real, dimension(:), allocatable :: countsource         ! source counts
    real, dimension(:), allocatable :: counterr            ! error on source counts
    real, dimension(:), allocatable :: sourceBackGroundCount  ! Source background counts
    real, dimension(:), allocatable :: psfwidth_x          ! width of psf
    real, dimension(:), allocatable :: psfwidth_y          ! width of psf
    real, dimension(:), allocatable :: poserr              ! error on position
    real, dimension(:), allocatable :: paSource
   
    integer(kind=int32), dimension(:), allocatable :: noSourcePixels ! The number of pixels included by a source
 
    integer(kind=int8), dimension(:), allocatable :: extendSource
    integer(kind=int8), dimension(:), allocatable :: confuseSource
  

    integer(kind=int32), pointer :: srcnum(:)  ! the source number
    real(kind=Single), pointer   :: r32(:)     ! pointer for list
    real(kind=Double), pointer   :: r64(:)     ! pointer for list
    integer(kind=Int8), pointer  :: i8(:)      ! pointer for list

    !***********************
    ! input image variables
    !***********************

    integer                      :: ndim     ! no of dimensions
    integer                      :: nx, ny   ! x,ysize of image
    integer                      :: dim(2)   ! image dimension array
    real                         :: exposure
    real(kind=Single), pointer   :: image(:, :)
    integer(kind=int16), pointer :: int16array(:, :)

  
    !************************************
    ! Background determination variables
    !************************************
    real               :: smoothSize   ! background smoothing size
    real               :: minDim

    !******************************
    ! source searching variables
    !******************************
    integer               :: nSource
    real, allocatable     :: background(:,:)
    real, allocatable     :: smBackground(:, :)
    real, dimension(:,:), pointer :: backgroundP
    integer(kind=int16), dimension(:, :), pointer :: levelImageP
    real                  :: medianBackground
    integer               :: i, j, m, k, l        ! loop counters
    integer               :: lPlus1
    real                  :: nSigma
    real                  :: fraction
    real                  :: minSigma
    real                  :: width

    character(len=1024)   :: outputBoxImage
    character(len=1024)   :: significanceMap

    integer(kind=int16), allocatable  :: levelImage(:, :)
    integer               :: allocErr
    integer               :: nSources
    integer               :: sourceConnect

    type(sourceParams), pointer :: sourcePars(:)
    integer               :: nSourcePars
    integer,allocatable :: minX(:), maxX(:), minY(:), maxY(:)
    logical(kind=bool) :: outputRegionFile
    integer :: ioUnit
    real(kind=double) :: plateScale, lambda

    real :: detectionRate
    real :: nullValue
    real (kind=double) :: apertureRadius     ! Radius (arcsecs) of circular aperture to be used for photometry
!
!-------------------------------------------------------------------------------------
!
    type(TableT)                                     :: inTableH     ! Input table handle.
    type(ColumnT), dimension(4)                      :: inColumnH    ! Input image column handle.

    integer                                          :: istat
    integer(kind=int32)                              :: blank

    integer(kind=int32)                              :: windowX0     ! WindowX0.
    integer(kind=int32)                              :: windowY0     ! WindowY0.
    integer(kind=int32)                              :: windowDX     ! WindowDX.
    integer(kind=int32)                              :: windowDY     ! WindowDY.

    integer(kind=int32), dimension(:), pointer       :: windowX0P    ! WindowX0 pointer.
    integer(kind=int32), dimension(:), pointer       :: windowY0P    ! WindowY0 pointer.
    integer(kind=int32), dimension(:), pointer       :: windowDXP    ! WindowDX pointer.
    integer(kind=int32), dimension(:), pointer       :: windowDYP    ! WindowDY pointer.

    integer(kind=int32)                              :: xMin         ! Minimum x value for window.
    integer(kind=int32)                              :: xMax         ! Maximum x value for window.
    integer(kind=int32)                              :: yMin         ! Minimum y value for window.
    integer(kind=int32)                              :: yMax         ! Maximum y value for window.

    type(ArrayT)                                     :: qualArrayH   ! Quality Array handle.
    integer(kind=int8), dimension(:, :), pointer     :: qualArrayP   ! Quality array pointer.
    logical(kind=bool), dimension(:, :), allocatable :: qualArray    ! Quality array.

    integer(kind=int32)                              :: filterIdInt


!
!--------------------------------------------------------------------------------------
!
 
    nullValue = -real(99999)
  

    !*************************
    ! get the task parameters
    !*************************
    minSigma = 2

    set = stringParameter("set")
    outset = stringParameter("outset")
    nSigma = realParameter("nsigma")
    fraction = realParameter("contrast")
    backgroundImage = stringParameter("background")
    significanceMap = stringParameter("signifimage")
    levelImageName = stringParameter("levelimage")
    smoothSize = realParameter("smoothsize")
    boxScale = intParameter("boxscale")  ! Now should be entered in arcsecs (12.12.2000)
                                    
    maxScale = intParameter("maxscale")
    maxScale = MAX(maxScale, 1)
    sourceConnect = intParameter("pixelconnect")
    outputBoxImage = stringParameter("boximage")
    flatField = stringParameter("flatset")
    mod8noise = stringParameter("mod8set")
    regionfile = stringParameter("regionfile")
    outputRegionFile = booleanParameter("outputregionfile")
    apertureRadius = realParameter("aperradius")

    !*************************
    ! get handle on osw image
    !*************************
    oswDataSet = DataSet(set,read)

    oswArray = array(oswDataSet, "PRIMARY", read)

    ndim  = numberOfDimensions(oswArray)             ! How many dimensions?

    ! If it's not an image then crash out

    if (ndim .ne. 2) call error('notImage','input file not an image')

    ! Get the exposure time in secs

    exposure = real32Attribute(oswDataSet,"EXPOSURE")

    dim = dimensions(oswArray)               ! Get the size of the image

    nx = dim(1)
    ny = dim(2)

    !**********************************************************************
    ! convert the image (useful for testing on DSS images, which are int16)
    !**********************************************************************
    if (arrayDataType(oswArray) .eq. Real32) then
        image => real32Array2Data(oswArray)

    else
        int16array => int16array2data(oswArray)
        allocate(image(0 : nx - 1, 0 : ny - 1), stat = allocErr)
        call CheckAllocation(allocErr, "image")
        image = real(int16array)
   

    end if


    !********************************************
    !
    ! AMK - 23/08/2000
    !*********************************************
    blank = int32Attribute(oswDataSet, "BLANK")

    !**********************************
    ! compute binning
    !**********************************
    binx = int32Attribute(oswDataSet, "BINAX1")
    biny = int32Attribute(oswDataSet, "BINAX2")
    twixel = booleanAttribute(oswDataSet, "BINBPE")

    omFilter = stringAttribute(oswDataSet, "FILTER")

    binx = 2 ** binx
    biny = 2 ** biny


    if (twixel) then
        binx = binx * 2
        biny = biny * 2
    end if

    if (binx .ne. biny) then
        call fatal("diffBin","Version 1 cannot deal with different binning factors in x and y")
    endif


    !**************************************************
    ! Get a handle on the modes binary table extension.
    !**************************************************
    inTableH = table(oswDataSet, "MODES")

    !**************************************************
    ! Get handles and pointers on the modes binary table 
    ! extension columns and extract the window data.
    !**************************************************
    inColumnH(1) = column(inTableH, "WINDOWX0", read)
    windowX0P => int32Data(inColumnH(1))
    windowX0 = windowX0P(0)

    inColumnH(2) = column(inTableH, "WINDOWY0", read)
    windowY0P => int32Data(inColumnH(2))
    windowY0 = windowY0P(0)

    inColumnH(3) = column(inTableH, "WINDOWDX", read)
    windowDXP => int32Data(inColumnH(3))
    windowDX = windowDXP(0)

    inColumnH(4) = column(inTableH, "WINDOWDY", read)
    windowDYP => int32Data(inColumnH(4))
    windowDY = windowDYP(0)

    xMin = windowX0
    xMax = windowX0 + windowDx - 1
    yMin = windowY0
    yMax = windowY0 + windowDy - 1
 
    !**************************************
    ! Allocate space for the quality array.
    !**************************************
    allocate (qualArray(1 : windowDX / binX, 1 : windowDY / binY), stat = allocErr)
    call CheckAllocation(allocErr, "qualArray")

    ! Get a handle and pointer on the quality array and extract the quality data.
    qualArrayH = array(oswDataSet, "QUALITY", read)
    qualArrayP => int8Array2Data(qualArrayH)

    qualArray = .false.
    where (qualArrayP .ne. 0) qualArray = .true.

    !*********************************
    !  Get the filter Id integer value
    !*********************************


    filterIdInt = filterIdToInt(omFilter)
   
    !**********************
    ! Set the CAL state
    !**********************
    call CAL_setState(OM, filterId = filterIdInt)

    
    !************************************* 
    ! Get the Plate scale (arcsecs/pixel)
    !*************************************
    call CAL_omGetPlateScale(lambda, plateScale)



    !****************************************
    ! Convert boxscale from arcsecs to pixels - CB-12-12-2000  
    !****************************************
    boxScale = ceiling(real(boxScale) / plateScale)
    

    !*************************************************************
    ! Set boxscale to take into account PSF of different filters
    ! 5 pixels minimum for optical filters (white,V,B FWHM ~ 2.6 pixels) 
    ! 7 pixels minimum for U+UV filters  (U,UV* FWHM ~ 3.8 pixels)
    !*************************************************************
    select case (omFilter)
        case ("V", "B", "WHITE", "white")
            boxScale = max(boxScale, 5)

        case ("U", "UVW1", "UVW2", "UVM2")
            boxScale = max(boxScale, 7)

         case default
             call fatal("WrongFilter","Filter in not one of White,WHITE,V,B,U,UVW1,UVW2,UVM2")

    end select



    !*********************************************************
    ! Take into account binning but make sure at least 3x3 box
    !*********************************************************
    boxScale = max(boxScale / binx, 3)

    write(errstr,'(a, i3)') "Minimum detection box set to : ", boxScale
    call message(AppMsg, Verbose)

    !************************************
    ! First smooth image on some scale 
    !************************************
    allocate(smBackground(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "smBackground")


    write(errstr,'(a)') "Beginning smoothing image (get median)"
    call message(AppMsg, Verbose)

    !********************************************************
    ! Smooth on small scale to get median value
    ! Needed if in poisson limit or low counts (almost always)
    !*********************************************************
    minDim = min(nx, ny) / 10.0
    smoothSize = min(smoothSize, minDim)
    smoothSize = max(smoothSize, 1.0)

    !********************************************
    ! image is the input image and smBackground
    ! the smoothed background image.
    !********************************************
    call smoothImage(image, nx, ny, smBackground, 3.0)
    write(errstr,'(a)') "Finished smoothing image "
    call message(AppMsg, Verbose)

    !**********************************
    ! get median of whole image
    !**********************************
    write(errstr,'(a)') "Getting medianBackground "
    call message(AppMsg, Verbose)
    call computeMode(smBackground, nx, ny, 3.0, medianBackground)


    write(errstr,'(a)') "Finished getting medianBackground "
    call message(AppMsg, Verbose)

    write(errstr,'(a, f7.1)') "Median of image: ", medianBackground
    call message(AppMsg, Verbose)

    if(medianBackground * boxScale * boxscale .lt. 15) then
        write(errstr,'(a, f7.1)') "Assuming Poisson statistics "
        call message(AppMsg, Verbose)
        poisson = .true.
    else
        write(errstr,'(a, f7.1)') "Assuming Gaussian statistics "
        call message(AppMsg, Verbose)
        poisson= .false.
    end if

    !***************************************************
    ! compute the background using a median box filter
    !***************************************************

    write(errstr,'(a)') "Computing the background"
    call message(AppMsg, Verbose)

    allocate(background(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "background")

    call computeBackground(image, nx, ny, background, smBackground, &
           medianBackground, 3.0)

    write(errstr,'(a)') "Beginning smoothing image (second time)"
    call message(AppMsg, Verbose)

    !**********************************************
    ! Make sure smooth size is less than image size
    !**********************************************
    minDim = min(nx, ny) / 10.0
    smoothSize = min(smoothSize, minDim)
    smoothSize = max(smoothSize, 1.0)
    call smoothImage(background, nx, ny, smBackground, smoothSize)
    write(errstr,'(a)') "Finished smoothing image (second time)"
    call message(AppMsg, Verbose)

    deallocate (background, stat = allocErr)
    call CheckDeallocation(allocErr, "background")

    write(errstr,'(a)') "Finished computing the background"
    call message(AppMsg, Verbose)

    !***********************************************
    ! Get numbers of detection limits
    !**********************************************
    call computeMode(smBackground, nx, ny, 3.0, medianBackground)

    write(errstr,'(a, f7.1)') "Smoothed median background : ", medianBackground
    call message(AppMsg, Verbose)

    !*****************************************
    ! Output background image if wanted
    !*****************************************
    if(backgroundImage .ne. "")then

        !*************************
        ! Create image file 
        !*************************
        backgroundSet = DataSet(backgroundImage, Create)

        backgroundArray = addArray(backgroundSet, "Background", Real32, &
                                   shape(smBackground))

        backgroundP => real32Array2Data(backgroundArray)

        backgroundP = smBackground

        call release(backgroundArray)
        call release(backgroundSet)

    end if

    !***************************
    ! Get significance Map
    !***************************
    write(errstr,'(a)') "Running source detection algorithm"
    call message(AppMsg, Verbose)

    allocate(levelImage(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "levelImage")

    levelImage = 0
    !*****************************************************
    ! Find sources in the image (image is the input image 
    ! and smBackground the smoothed image.
    ! levelImage is an image showing the initially 
    ! detected sources (Each pixel of a source has the
    ! pixel value set to the source number
    !*****************************************************
    call sourceDetectManyScales(image, nx, ny, smBackground, nSigma, minSigma, &
         boxScale, maxScale, nSources, levelImage, significanceMap, poisson)

    write(errstr,'(a,i5)') &
             "Finished running source detection algorithm - nsources: ",nSources
    call message(AppMsg, Verbose)
    write(errstr,'(''Nsources = '', i5)') nSources
    call message(AppMsg, Verbose)

    !*******************************************************
    ! Level image has all the information we need...
    !*******************************************************
    if(levelImageName .ne. "")then

        !***************************
        ! Create image file 
        !*************************** 
        levelImageSet = DataSet(levelImageName, Create)

        levelImageArray = addArray(levelImageSet, "Background", Integer16, &
                 shape(levelImage))

        levelImageP => int16Array2Data(levelImageArray)

        levelImageP = levelImage

        call release(levelImageArray)
        call release(levelImageSet)

    end if

    !*******************************************************
    ! Get min/max values for all sources from levelImage
    !******************************************************
    allocate(minX(nSources), stat = allocErr)
    call CheckAllocation(allocErr, "minX")

    allocate(maxX(nSources), stat = allocErr)
    call CheckAllocation(allocErr, "maxX")

    allocate(minY(nSources), stat = allocErr)
    call CheckAllocation(allocErr, "minY")

    allocate(maxY(nSources), stat = allocErr)
    call CheckAllocation(allocErr, "maxY")

    call getMinMaxFromLevelImage(levelImage, nx, ny, nSources, minX, maxX, minY, maxY)


    !*******************************************************
    ! Allocate memory for arrays that will store the source 
    ! parameters for writing to the output fits file
    !*******************************************************
    size = max(100, nSources * size) 
    allocate(xsource(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "xsource")

    allocate(ysource(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "ysource")

    allocate(sourcesignif(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "sourcesignif")

    allocate(countsource(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "countsource")

    allocate(counterr(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "counterr")
  
    allocate(sourceBackGroundCount(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "sourceBackGroundCount")
 
    allocate(psfwidth_x(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "psfwidth_x")

    allocate(psfwidth_y(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "psfwidth_y")

    allocate(poserr(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "poserr")
    
    allocate(paSource(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "paSource")

    allocate(extendSource(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "extendSource")

    allocate(confuseSource(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "confuseSource")

    allocate(noSourcePixels(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "noSourcePixels")

    !************************************************************
    ! for each of the sources get sources even if blended
    !************************************************************
    k = 0
    do i = 1, nSources
        write(errstr, '(''Analysing source '', i5)') i
        call message(AppMsg, Verbose)

        call detectParams(nx, ny, image, smBackground, levelImage, i, &
             minX(i), maxX(i), minY(i), maxY(i), nSigma, fraction, &
             nSourcePars, sourcePars)
!        write(*, *) "No of sources=", nSources, " no of source pars=", nSourcePars


        if(nSourcePars .gt. 0) then

            !**************************************
            ! remove identical sources if any
            !**************************************
            j = 0
            removeMultiple: do
                j = j + 1
                if(j .gt. nSourcePars) exit removeMultiple
                m = j
                removeMultiple2: do 
                    m = m + 1
                    if(m .gt. nSourcePars) exit removeMultiple2
                    if((sourcePars(m)%x .eq. sourcePars(j)%x) .and. &
                       (sourcePars(m)%y .eq. sourcePars(j)%y) .and. &
                       (sourcePars(m)%semiMajor .eq. sourcePars(j)%semiMajor) .and. &
                       (sourcePars(m)%semiMinor .eq. sourcePars(j)%semiMinor)) then

                        forall (l = m : nSourcePars - 1)
                            lPlus1 = l + 1
                            sourcePars(l) = sourceParams( &
                                            sourcePars(lPlus1)%x, sourcePars(lPlus1)%y, &
                                            sourcePars(lPlus1)%errX, sourcePars(lPlus1)%errY, &
                                            sourcePars(lPlus1)%PosAngle, sourcePars(lPlus1)%semiMajor, &
                                            sourcePars(lPlus1)%semiMinor, sourcePars(lPlus1)%countrate, &
                                            sourcePars(lPlus1)%countrateErr, sourcePars(lPlus1)%backGroundRate, &
                                            sourcePars(lPlus1)%noSourcePixels, sourcePars(lPlus1)%signif, &
                                            sourcePars(lPlus1)%Cflag)
                        end forall

                        nSourcePars = nSourcePars - 1
                        m = m - 1
                    end if
                end do removeMultiple2
            end do removeMultiple

            do j = 1, nSourcePars
                if(sourcePars(j)%countrate .gt. 0) then
                    k = k + 1
                    xSource(k) = sourcePars(j)%x
                    ySource(k) = sourcePars(j)%y

                    !********************************************************
                    ! Set lower bound to 0.5 pixel for positional accuracy
                    !********************************************************

                    posErr(k) = max(max(sourcePars(j)%errX, sourcePars(j)%errY), 0.5) * &
                                max(binx, biny) * plateScale
                    paSource(k) = sourcePars(j)%PosAngle
                    countSource(k) = sourcePars(j)%countrate
                    countErr(k) = sourcePars(j)%countrateErr
                    sourceBackGroundCount(k) = sourcePars(j)%backGroundRate
                    psfwidth_x(k) = sourcePars(j)%semiMajor * binx * plateScale
                    psfwidth_y(k) = sourcePars(j)%semiMinor * biny * plateScale
                    sourceSignif(k) = sourcePars(j)%signif      
                    noSourcePixels(k) = sourcePars(j)%noSourcePixels

                    !********************
                    ! Set up flags
                    !********************
                    width = max(psfwidth_x(k), psfwidth_y(k))
                    !********************************************************
                    ! extended source flag if max width > 2 this is in sigma 
                    !********************************************************
                    if (width .gt. 3.0) then  
                        extendSource(k) = 1
                    else
                        extendSource(k) = 0
                    end if
            
                    confuseSource(k) = sourcePars(j)%Cflag            
                end if
            end do

        endif
    end do   ! End of loop through all the sources


    nSource = k
    if(allocated(minX)) then
        deallocate(minX, stat = allocErr)
        call CheckDeallocation(allocErr, "minX")
    end if

    if(allocated(maxX)) then
        deallocate(maxX, stat = allocErr)
        call CheckDeallocation(allocErr, "maxX")
    end if

    if(allocated(minY)) then
        deallocate(minY, stat = allocErr)
        call CheckDeallocation(allocErr, "minY")
    end if

    if(allocated(maxY)) then
        deallocate(maxY, stat = allocErr)
        call CheckDeallocation(allocErr, "maxY")
    end if

  
    !**********************************
    ! Output to saoiamge region file
    !**********************************
    if(outputRegionFile) then
        call get_free_io_unit(ioUnit)
        open(unit = ioUnit, file = regionfile, status = 'unknown')
        do i = 1, nSource
            write(ioUnit, '('' ELLIPSE('', f7.2,'', '', f7.2, '', '', f6.2, '', '', f6.2, '', '', f6.2, '')'')') &
                  xSource(i), ySource(i), psfwidth_x(i) * 3 / (binx * plateScale), &
                  psfwidth_y(i) * 3 / (biny * plateScale), paSource(i)
        end do
        close(ioUnit)
    end if

    !*********************************************
    ! Write out the source list to the oswDataSet
    !*********************************************
    write(errstr,'(a)') "Writing source list"
    call message(AppMsg, Verbose)

    !**********************************
    ! Create the output source list
    !**********************************
    listDataSet = dataSet(outset, create)

    listTable = addTable(listDataSet, "SRCLIST", nsource)

    !****************************************
    ! copy the keywords from the image header
    !****************************************

    call copyAllAttributes(attributable(oswDataSet), attributable(listDataSet))

    !*****************************
    ! Add in detection count rate
    ! which will be stored in the source list file
    ! and used by ommag to calculate the
    ! background limiting  magnitude
    !************************************

  
    apertureRadius = CAL_omGetApertureRadius() / real(binx)  

    if (medianBackground .gt. 0.0) then  ! was boxScale * boxScale- now the aperture area
        detectionRate = nSigma * sqrt(medianBackground * PI * apertureRadius * apertureRadius ) / exposure
    else
        detectionRate = 0.0
    end if 

    write(errstr,'(a, 1pg9.3)') "Detection Count Rate: ", detectionRate
    call message(AppMsg, Verbose)


    call setAttribute(listDataSet, "CRLIMIT", detectionRate, "counts/sec", "Detection limit count rate")


   
    ! this is an osw source list

    call setAttribute(listDataSet, "CONTENT", "OM OSW SOURCE LIST", &
                  "File contents")
    call setAttribute(listDataSet, "HDUCLASS", "OGIP", &
                  "Format conforms to OGIP/GSFC conventions")
    call setAttribute(listDataSet, "HDUCLAS1", "SRCLIST", &
                  "File contains a source list")

    !********************************
    ! now add the columns one by one
    ! starting with the source number
    !*********************************
    pcolumn = addColumn(listTable, "SRCNUM", Integer32, " ")
    srcnum => int32data(pcolumn)

    forall (i = 0 : nsource - 1) srcnum(i) = i +1
   
    call release(pcolumn)

    !*********************************
    ! Add the source positions
    !*********************************

    call AddReal32Column(listTable, "XPOS", "pixels", nsource, xsource, nullValue)
    call AddReal32Column(listTable, "YPOS", "pixels", nsource, ysource, nullValue)


    pcolumn = addColumn(listTable, "RA", Real64, "degrees")
    r64 => real64data(pcolumn)
    forall (i = 0 : nsource - 1) r64(i) = 0.0

    call release(pcolumn)


    pcolumn = addColumn(listTable, "DEC", Real64, "degrees")
    r64 => real64data(pcolumn)
    forall (i = 0 : nsource - 1) r64(i) = 0.0
    
    call release(pcolumn)

    call message(AppMsg, Verbose)

 

    call AddReal32Column(listTable, "POSERR", "pixels", nsource, poserr, nullValue)

    !*********************************************
    ! Add the source-count rates
    !*********************************************

    countsource = countsource / exposure
    call AddReal32Column(listTable, "RATE", "counts/sec", nsource, countsource, nullValue)

    counterr = counterr / exposure
    call AddReal32Column(listTable, "RATE_ERR", "counts/sec", nsource, counterr, nullValue)

    !*********************************************
    ! The source background rate (Added by CB)
    !*********************************************
    sourceBackGroundCount = sourceBackGroundCount / exposure
    call AddReal32Column(listTable, "BACKGROUND_RATE", "counts/sec", nsource, sourceBackGroundCount, nullValue)

    call AddReal32Column(listTable, "SIGNIFICANCE", "sigma", nsource, sourcesignif, nullValue)

    !************************************
    ! Add the magnitudes (set to zero!)
    !************************************ 
    xsource = 0.0
    call AddReal32Column(listTable, "MAG", "magnitudes", nsource, xsource, nullValue)
    call AddReal32Column(listTable, "MAGERR", "magnitudes", nsource, xsource, nullValue)



    pcolumn = addColumn(listTable, "FWHM_MAJ", Real32, "pixels")
    r32 => real32data(pcolumn)

    forall (i = 0 : nsource - 1) r32(i) = max(2.354 * psfwidth_x(i + 1), 2.354 * psfwidth_y(i + 1))
    
    call release(pcolumn)


    pcolumn = addColumn(listTable, "FWHM_MIN", Real32, "pixels")
    r32 => real32data(pcolumn)

    forall (i = 0 : nsource - 1) r32(i) = min(psfwidth_x(i + 1) * 2.354, psfwidth_y(i + 1) * 2.354)
  
    call release(pcolumn)


    call AddReal32Column(listTable, "PA", "degrees", nsource, paSource, nullValue)

    !**********************************
    ! Add the quality flags
    !**********************************
    pcolumn = addColumn(listTable, "QFLAG", Integer8, "none")
    i8 => int8data(pcolumn)
    i8 = 0.0
    call release(pcolumn)

    call AddInt8Column(listTable, "CFLAG", "none", nsource, confuseSource)
    call AddInt8Column(listTable, "EFLAG", "none", nsource, extendSource)

    !************************
    ! release the memory
    !************************
    call release(oswArray)

    call release(listTable)
    call release(listDataSet)
    call release(oswDataSet)

    deallocate(nosourcePixels, stat = allocErr)
    call CheckDeallocation(allocErr, "noSourcePixels")


    deallocate(xsource, stat = allocErr)
    call CheckDeallocation(allocErr, "xsource")


    deallocate(ysource, stat = allocErr)
    call CheckDeallocation(allocErr, "ysource")


    deallocate(sourcesignif, stat = allocErr)
    call CheckDeallocation(allocErr, "sourcesignif")


    deallocate(countsource, stat = allocErr)
    call CheckDeallocation(allocErr, "countsource")


    deallocate(counterr, stat = allocErr)
    call CheckDeallocation(allocErr, "counterr")

  
    deallocate(sourceBackGroundCount, stat = allocErr)
    call CheckDeallocation(allocErr, "sourceBackGroundCount")

 
    deallocate(psfwidth_x, stat = allocErr)
    call CheckDeallocation(allocErr, "psfwidth_x")

    deallocate(psfwidth_y, stat = allocErr)
    call CheckDeallocation(allocErr, "psfwidth_y")

      
    deallocate(poserr, stat = allocErr)
    call CheckDeallocation(allocErr, "poserr")
 
   
    deallocate(paSource,  stat = allocErr)
    call CheckdeAllocation(allocErr, "paSource")

    deallocate(extendSource,  stat = allocErr)
    call CheckdeAllocation(allocErr, "extendSource")

    deallocate(confuseSource,  stat = allocErr)
    call CheckdeAllocation(allocErr, "confuseSource")

    if(allocated(qualArray)) then
        deallocate(qualArray, stat = istat)
        call CheckDeallocation(istat, "qualArray")
    end if

    if(allocated(levelImage)) then
        deallocate(levelImage, stat = istat)
        call CheckDeallocation(istat, "levelImage")
    end if




!
!---------------------------------------------------------------------
!

! that's all
!    if (noAllocations .ne. noDeallocations) then
!        write(*, *) "No of allocations = ", noAllocations, " noDeallocations=", noDeallocations
!    end if

write(errstr,'(a)') "Exiting..."
call message(AppMsg, Verbose)




end subroutine omdetect

!******************************************************************
! Get min and max values for each source from levelImage
!
!******************************************************************
subroutine getMinMaxFromLevelImage(levelImage, nx, ny, nSources, &
           minX, maxX, minY, maxY)

    integer, intent(in) :: nx, ny
    integer(kind=int16), intent(in) :: levelImage(0 : nx - 1, 0 : ny - 1)
    integer, intent(in) :: nSources
    integer, intent(out) :: minX(nSources)
    integer, intent(out) :: maxX(nSources)
    integer, intent(out) :: minY(nSources)
    integer, intent(out) :: maxY(nSources)

    integer :: i, j
    integer :: levelImageVal

!print *,"getMinMaxFromLevelImage"

    forall (i = 1 : nSources)
        minX(i) = nx + 1
        maxX(i) = 0
        minY(i) = ny + 1
        maxY(i) = 0
    end forall

    do i = 0, ny - 1
        do j = 0, nx - 1
            levelImageVal = levelImage(j, i)

            if(levelImageVal .ne. 0) then
                minX(levelImageVal) = MIN(minX(levelImageVal), j + 1)
                maxX(levelImageVal) = MAX(maxX(levelImageVal), j + 1)
                minY(levelImageVal) = MIN(minY(levelImageVal), i + 1)
                maxY(levelImageVal) = MAX(maxY(levelImageVal), i + 1)
            end if
        end do
    end do

end subroutine getMinMaxFromLevelImage
!***************************************************************
! subroutine nextToPosition
!
!***************************************************************
recursive subroutine nextToPosition(newImage, levelImage, nx, ny, xpos, ypos, ok)

    integer, intent(in) :: nx, ny
    real, intent(inout) :: newImage(nx, ny)
    real, intent(inout) :: levelImage(nx, ny)
    integer, intent(in) :: xpos, ypos
    logical, intent(out) :: ok

    integer :: minXpos, maxXpos, minYpos, maxYpos
    integer :: i, j

!  print *,"nextToPosition"
  
    ok = .false.

    minXpos = MAX(1, xpos - 1)
    minYpos = MAX(1, ypos - 1)
    maxXpos = MIN(nx, xpos + 1)
    maxYpos = MIN(ny, ypos + 1)
    do i = minYpos, maxYpos
        do j = minXpos, maxXpos
            if(levelImage(j, i) .ne. 0) then
                if(i .ne. xpos .and. j .ne. ypos) then
                    if(newImage(j, i) .gt. 0) then
                        levelImage(xpos, ypos) = levelImage(j, i)
                        newImage(xpos, ypos) = 0
                        call nextToPosition(newImage, levelImage, nx, ny, j, i, ok)
                        ok = .true.
                    end if
                end if
            end if
        end do
    end do


end subroutine nextToPosition

!************************************************************
! subroutine sourceDetectManyScales
! Purpose: Detects all the sources in the image "image"- some may be
! detected twice.
! Returns: The number of sources detected "nSources" and the image
! "levelImage" - 
! image is the input image
! nx & ny the image dimensions
! background is the smoothed background image
!************************************************************
subroutine sourceDetectManyScales(image, nx, ny, background, nSigma, minSigma, &
     boxScale, maxScale, nSources, levelImage, significanceMap, poisson)

    integer, intent(in) :: nx, ny
    real, intent(in) :: image(0 : nx - 1, 0 : ny - 1)
    real, intent(in) :: background(0 : nx - 1, 0 : ny - 1)
    real, intent(in) :: nSigma, minSigma
    integer, intent(in) :: boxScale
    integer, intent(in) :: maxScale
    integer, intent(out) :: nSources
    integer(kind=int16), intent(out) :: levelImage(0 : nx - 1, 0 : ny - 1)
    character(len=*), intent(in) :: significanceMap
    logical(kind=bool), intent(in) :: poisson
!    integer, intent(in) :: sourceConnect

    type(DataSetT)       :: significanceSet 
    type(ArrayT)         :: significanceArray         
    real, dimension(: , :), pointer :: significanceP
  
    integer :: origNx, origNy
    integer :: oldNx, oldNy
    integer :: newNx, newNy
    integer :: nSource

    integer :: allocErr
    real, allocatable :: sigmaMap(:, :)
    real, allocatable :: oldSigmaMap(:, :)
    real, allocatable :: inImage(:, :)
    real, allocatable :: inBackground(:, :)
    real, allocatable :: newImage(:, :)
    real, allocatable :: newBackground(:, :)

    real :: sigmaMapThreshold
    real :: minSigmaMapThreshold

    integer :: scale

    !*************************************
    ! Do things at first scale 
    !*************************************
    allocate(oldSigmaMap(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "oldSigmaMap")

  ! Allocatable arrays of variable size - start at level 1

    !*************************************
    ! Do it at first scale
    !*************************************
    write(errstr, '(a)') "Running source detection algorithm (level 1)"
    call message(AppMsg, Verbose)  ! Verbose was 5

    call sourceDetect(image, nx, ny, background, oldSigmaMap, &
       sigmaMapThreshold, minSigmaMapThreshold, nSigma,minSigma, boxScale, &
       poisson)

    write(errstr,'(a)') "Finished running source detection algorithm (level 1)"
    call message(AppMsg, Verbose)  ! Verbose was 5

    !**********************************
    ! Output significance map if wanted
    !***********************************
    if(significanceMap .ne. "") then


        !**********************
        ! Create image file 
        !**********************
        significanceSet = DataSet(significanceMap, Create)

        significanceArray = addArray(significanceSet, "Signifiance Map", Real32, &
                            shape(oldSigmaMap))
     
        significanceP => real32Array2Data(significanceArray)
     
        significanceP = oldSigmaMap
     
        call release(significanceArray)
        call release(significanceSet)
     
    end if

    !**************************************
    ! Now start to make level image
    !**************************************
    write(errstr,'(a)') "Getting source number image (level 1)"
    call message(AppMsg, Verbose)  ! Verbose was 5

    nSource = 0

    allocate(sigmaMap(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "sigmaMap")
    sigmaMap = oldSigmaMap

    print *,"SigmaMapThreshold:", sigmaMapThreshold
    call setSources(sigmaMap, nx, ny, 1, oldSigmaMap, levelImage, nx, ny, &
               sigmaMapThreshold, minSigmaMapThreshold, nSource)

    if(allocated(sigmaMap)) then
        deallocate(sigmaMap, stat = allocErr)
        call CheckDeallocation(allocErr, "sigmaMap")
    end if
  
    write(errstr, '(a)') "Finished getting source number image (level 1)"
    call message(AppMsg, Verbose)  ! Verbose was 5
 
    origNx = nx
    origNy = ny

    ! Now go down in levels (ie. rebinning of image)

    oldNx = nx
    oldNy = ny
    allocate(inImage(0 : oldNx - 1, 0 : oldNy - 1), stat = allocErr)
    call CheckAllocation(allocErr, "inImage")
    inImage = image
    allocate(inBackground(0 : oldNx - 1, 0 : oldNy - 1), stat = allocErr)
    call CheckAllocation(allocErr, "inBackground")


    inBackground = background

    scale = 1
    scaleLoop: do

        scale = scale + 1
        if(scale .gt. maxScale) exit scaleLoop
        

        write(errstr,'(a,i5)') "Analysing image a at level ",scale
        call message(AppMsg, Verbose) ! Verbose was 5

        newNx = oldNx / 2
        newNy = oldNy / 2

        ! First - don't want to get tiny image!

        if(newNx .lt. 32 .or. newNy .lt. 32) exit scaleLoop
       

        allocate(newImage(0 : newNx - 1, 0 : newNy - 1), stat = allocErr)
        call CheckAllocation(allocErr, "newImage (sourceDetectManyScalee")

        allocate(newBackground(0 : newNx - 1, 0 : newNy - 1), stat = allocErr)
        call CheckAllocation(allocErr, "newBackground")

        !***********************************
        ! Rebin images by factor of 2
        !***********************************
        write(errstr, '(a, i5)') "Rebinning image at level ", scale
        call message(AppMsg, Verbose)  ! Verbose was 5

        call rebinImageBy2(inImage, oldNx, oldNy, newImage, newNx, newNy)
        call rebinImageBy2(inBackground, oldNx, oldNy, newBackground, newNx, newNy)

        ! Allocate sigmaMap

        allocate(sigmaMap(0 : newNx - 1, 0 : newNy - 1), stat = allocErr)
        call CheckAllocation(allocErr, "sigmaMap")
    
        !********************
        ! Source detect
        !********************

        write(errstr, '(a, i5)') "Source detect at level ", scale
        call message(AppMsg, Verbose)  ! Verbose was 5
        call sourceDetect(newImage, newNx, newNy, newBackground, sigmaMap, &
                          sigmaMapThreshold, minSigmaMapThreshold, nSigma, &
                          minSigma, boxScale, poisson)
        !*****************************
        ! add to levelImage map
        !*****************************
        call setSources(sigmaMap, newNx, newNy, scale, oldSigmaMap, levelImage, origNx, &
                        origNy, sigmaMapThreshold, minSigmaMapThreshold, nSource)

        if(allocated(sigmaMap)) then
            deallocate(sigmaMap, stat = allocErr)
            call CheckDeallocation(allocErr, "sigmaMap")
        end if

        write(errstr, '(a, i4)') "Finished getting source number image at scale ", scale
        call message(AppMsg, Verbose)  ! Verbose was 5


        ! Tidy up image arrays

        if(allocated(inImage)) then
            deallocate(inImage, stat = allocErr)
            call CheckDeallocation(allocErr, "inImage")
        end if

        if(allocated(inBackground)) then
            deallocate(inBackground,stat = allocErr)
            call CheckDeallocation(allocErr, "inBackground")
        end if

        allocate(inImage(0 : newNx - 1, 0 : newNy - 1), stat = allocErr)
        call CheckAllocation(allocErr, "inImage")

        inImage = newImage
        allocate(inBackground(0 : newNx - 1, 0 : newNy - 1), stat = allocErr)
        call CheckAllocation(allocErr, "inBackground (sourceDetectManyScales")

        inBackground = newBackground

        if(allocated(newimage)) then
            deallocate(newImage, stat = allocErr)
            call CheckDeallocation(allocErr, "newIMage")
       end if

        if(allocated(newBackground)) then
            deallocate(newBackground,stat=allocErr)
            call CheckAllocation(allocErr, "newBackground")
       end if

        oldNx = newNx
        oldNy = newNy

    end do scaleLoop

    if(allocated(inImage)) then
        deallocate(inImage, stat = allocErr)
        call CheckDeallocation(allocErr, "inImage")
    end if

    if(allocated(inBackground)) then
        deallocate(inBackground, stat = allocErr)
        call CheckDeallocation(allocErr, "inBackground")
    end if

    if(allocated(oldSigmaMap)) then
        deallocate(oldSigmaMap, stat = allocErr)
        call CheckDeallocation(allocErr, "oldSigmaMap")
    end if

    !***************************************************
    ! Get nSources from the maximum value  in levelImage
    !***************************************************
    nSources = MAXVAL(levelImage)

end subroutine sourceDetectManyScales

!****************************************************************
! subroutine rebinImageBy2 
!
!****************************************************************
subroutine rebinImageBy2(inImage, oldNx, oldNy, newImage, newNx, newNy)
 
    integer, intent(in) :: oldNx, oldNy
    real, intent(in) :: inImage(0 : oldNx - 1, 0 : oldNy - 1)
    integer, intent(in) :: newNx, newNy
    real, intent(out) :: newImage(0 : newNx - 1, 0 : newNy - 1)

    integer :: i, j, k, l
    real :: newImageVal

!  print *,"rebinImageBy2"

    if((oldNx / 2 .ne. newNx) .or. (oldNy / 2 .ne. newNy)) then
     call fatal("rebinImageBy2","Input arrays not factors of 2")
  endif

  forall (i = 0 : newNy - 1)
     k = i * 2
     forall (j = 0 : newNx - 1)
        l = j * 2
        newImageVal = inImage(l, k)
        newImageVal = newImageVal + inImage(l + 1, k)
        newImageVal = newImageVal + inImage(l, k + 1)
        newImageVal = newImageVal + inImage(l + 1, k + 1)
        newImage(j, i) = newImageVal
     end forall
  end forall

end subroutine rebinImageBy2
!****************************************************
! subroutine sourceDetect
! Sliding box algorithm for source detection
!****************************************************

subroutine sourceDetect(image, nx, ny, background, sigmaMap, &
     sigmaMapThreshold, minSigmaMapThreshold, nSigma, minThreshold, &
     boxScale, poisson)
 
    integer, intent(in) :: nx, ny
    real, intent(in) :: image(0 : nx - 1, 0 : ny - 1)
    real, intent(in) :: background(0 : nx - 1, 0 : ny - 1)
    real, intent(inout) :: sigmaMap(0 : nx - 1, 0 : ny - 1)
    real, intent(inout) :: sigmaMapThreshold
    real, intent(inout) :: minSigmaMapThreshold
    real, intent(in) :: nSigma
    real, intent(in) :: minThreshold
    integer, intent(in) :: boxScale
    logical(kind=bool), intent(in) :: poisson

  ! local variables

    real, allocatable   :: scan(:)
    integer, allocatable :: nscan(:)

    real, allocatable :: newBAckground(:, :)
    real, allocatable :: boxImage(:, :)

    integer :: i, j
    integer :: height
    real :: backgroundVal, sigmaMapVal
    integer :: imageVal
    integer :: allocErr

!  print *[B,"SourceDetect"

    height = boxScale / 2

    allocate(scan(-height : height), stat = allocErr)
    call CheckAllocation(allocErr, "scan (sourceDetect)")


    allocate(nscan(-height : height), stat = allocErr)
    call CheckAllocation(allocErr, "nscan (sourceDetect)")

    allocate(boxImage(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "boxImage (sourceDetect)")

    allocate(newBackground(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "newBackground (sourceDetect)")

    !************************************     
    ! Slide the box to get binned image
    !************************************
    do i = 0, nx - 1
        call MovingBox(image, nx, ny, boxImage, height, height, i, scan, nscan, 2)
    end do

    !******************************************************
    ! Then background map (mod-8 noise on small scales!)
    !******************************************************

    do i = 0, nx - 1
        call MovingBox(background, nx, ny, newBackground, height, height, i, &
                       scan, nscan, 0)
    end do

    if(.not. poisson) then

        sigmaMapThreshold = nSigma
        minSigmaMapThreshold = minThreshold
     
        sigmaMap = (boxImage - newBackground) / sqrt(newBackground)

    else
        !******************************************************************
        ! Compute where data is above significance using Poisson statistics
        ! Calculate image as sigma not probability
        !******************************************************************
     
        sigmaMapThreshold = -log(newerrf(nSigma))
        minSigmaMapThreshold = -log(newerrf(minThreshold))

        do i = 0, ny - 1
            do j = 0, nx - 1
                backgroundVal = newBackground(j, i)
                imageVal = nint(boxImage(j, i))
           
                ! Get prob in log space
           
                if(imageVal .gt. backgroundVal) then
                    sigmaMapVal = imageVal * log(backgroundVal) - backgroundVal - &
                                  lnFactrl(imageVal)
                else
                    sigmaMapVal = 0.0
                end if
           
                sigmaMap(j, i) = -sigmaMapVal
            end do
        end do
      
    end if

    if(allocated(newBackground)) then
        deallocate(newBackground, stat = allocErr)
        call CheckDeallocation(allocErr, "newBackground")
    end if

    if(allocated(boxImage)) then
        deallocate(boxImage, stat = allocErr)
        call CheckDeallocation(allocErr, "boxImage")
    end if

    if(allocated(scan)) then
        deallocate(scan, stat = allocErr)
        call CheckDeallocation(allocErr, "scan")

    end if

    if(allocated(nscan)) then
        deallocate(nscan, stat = allocErr)
        call CheckDeallocation(allocErr, "nscan")
    end if

end subroutine sourceDetect

!***********************************************************************
!
! subroutine setSources - use 8 connectivity to match pixels
! fill in an image of where the sources are - the used to constrcut pixel
! lists to match output needed for TJH version of omdetect
!***********************************************************************
subroutine setSources(image, nx, ny, scale, fullImage, levelImage, oldNx, oldNy, &
     nSigma, minThreshold, nSources)

    !*********************************************************
    ! Sets up levelImage which contains only connected pixels
    ! Basically sets up where the sources are
    !*********************************************************
    integer, intent(in) :: nx,ny
    real, intent(inout) :: image(0 : nx - 1, 0 : ny - 1)
    integer, intent(in) :: scale
    integer, intent(in) :: oldNx, oldNy
    real, intent(in) :: fullImage(0 : oldNx - 1,0 : oldNy - 1)
    integer(kind=int16), intent(inout) :: levelImage(0 : oldNx - 1, 0 : oldNy - 1)
    real, intent(in) :: nSigma
    real, intent(in) :: minThreshold
    integer, intent(inout) :: nSources

    type(src), pointer :: sourceList
    integer :: nSourceList

    integer :: nPixel
    type(src), pointer :: currentSource
    type(srcpixel), pointer :: currentPixel
  
    integer :: minXpos, maxXpos
    integer :: minYpos, maxYpos

    integer :: i, j
    integer :: sourceNumber
    integer :: xpos, ypos
    integer :: numberSources
    integer :: scaling

!  print *,"setSources"

    nullify(sourceList) ! source list for this scale
    nSourceList = 0
  
    !*****************************************************
    ! At this scale we get linked list of source positions
    !*****************************************************
    write(errstr,'(a, i4)') "Getting linked sources at scale ", scale
    call message(AppMsg, Verbose)  ! Verbose was 5

    call getLinkedSources(image, nx, ny, nSigma, minThreshold, sourceList, nSourceList)

    write(errstr,'(a, i4)') "Finished getting linked sources at scale ", scale
    call message(AppMsg, Verbose)  ! Verbose was 5

    if(nSourceList .eq. 0) then
        return
    end if
    !************************************************************************
    ! Check source lists with image/linkedImage to make sure new sources are 
    ! not repeats of sources found at a higher resolution
    !************************************************************************
    currentSource => sourceList
    sourceLoop: do
     
        if(.not. associated(currentSource)) then
        exit sourceLoop
     else
        sourceNumber = currentSource%sourceNum
        nPixel = currentSource%numPixels
        if(nPixel .ge. 0) then

           ! data stored

           if(scale .eq. 1) then  ! at minimum scale

              nSources = max(nSources, sourceNumber)
              currentPixel => currentSource%pixelList
              pixelLoop: do 
                 if(.not. associated(currentPixel)) then
                    exit pixelLoop
                 else
                    xpos = currentPixel%x
                    ypos = currentPixel%y
                          
                    ! Not at multiple scales so just pixel/pixel

                    if(levelImage(xpos, ypos) .eq. 0) then
                       if(fullImage(xpos, ypos) .gt. minThreshold) then
                          levelImage(xpos, ypos) = sourceNumber
                       end if
                    end if
                 end if

                 ! Onto next pixel

                 currentPixel => currentPixel%next

              end do pixelLoop

           else

              numberSources = 0
              currentPixel => currentSource%pixelList
              pixelLoop2: do 
                 if(.not. associated(currentPixel)) then
                    exit pixelLoop2
                 else
                    xpos = currentPixel%x
                    ypos = currentPixel%y
                    
                    ! See if this source overlaps with other sources
                    
                    scaling = 2 ** (scale-1)
                    minYpos = ypos*scaling
                    maxYpos = minYpos + scaling - 1
                    minXpos = xpos*scaling
                    maxXpos = minXpos + scaling - 1
                    
! If there are sources already then ignore - already go down to minTresh
! at searching stage

                    countSources: do i=minYpos,maxYpos
                       
                       do j = minXpos, maxXpos
                          
                          if(levelImage(j, i) .gt. 0)then
                             numberSources = -1
                             exit countSources
                          end if
                          
                       end do
                       
                    end do CountSources

                 end if

                 ! Onto next pixel

                 currentPixel => currentPixel%next

              end do pixelLoop2

              ! Now know if this source is clear of already detected sources
              ! If we don't have sources then assign pixels
                    
              if(numberSources .ge. 0) then ! not multiple sources
                       
                 ! If new sources then increment number of sources
                       
                 if(numberSources .eq. 0) then
                    nSources = nSources + 1
                    numberSources = nSources
                 endif
                 
                 currentPixel => currentSource%pixelList
                 pixelLoop3: do 
                    if(.not. associated(currentPixel))then
                       exit pixelLoop3
                    else
                       xpos = currentPixel%x
                       ypos = currentPixel%y
                       
                       ! See if this source overlaps with other sources
                    
                       scaling = 2 ** (scale - 1)
                       minYpos = ypos * scaling
                       maxYpos = minYpos + scaling - 1
                       minXpos = xpos * scaling
                       maxXpos = minXpos + scaling - 1
                       
                       
                       ! Now assign values and add pixels to linked list
                       
                       do i = minYpos, maxYpos
                          do j = minXpos, maxXpos
                             
                             ! Test to see if we've been here before
                             ! and if we're in a detected pixel possibility
                             
                             if(levelImage(j, i) .eq. 0) then
                                if(fullImage(j, i) > minThreshold)then
                                   
                                   ! Is this connected to any other pixel?
                                   ! Use nearest neighbour again at this level
                                   ! within the range allowed by detection
                                   
                                   call nearestNeighbourImage(fullImage, &
                                        oldNx, oldNy, levelImage, minXpos, &
                                        maxXpos, minYpos, maxYpos, j, i, &
                                        minThreshold, numberSources)
                                   
                                end if
                             end if
                          end do
                          
                       end do

                       ! Onto next pixel

                       currentPixel => currentPixel%next

                    end if

                 end do pixelLoop3

              end if

           end if

        end if

        ! Onto next source

        currentSource => currentSource%next

     endif

  end do sourceLoop

  call freeSourceList(sourceList)

end subroutine setSources
!*****************************************
! subroutine freeSourceList
! Routine to free memory from sourceList
!*****************************************
subroutine freeSourceList(sourceList)

  type(src), pointer :: sourceList
  type(src), pointer :: currentSource
  type(srcPixel), pointer :: currentPixel

  integer             :: allocErr

!  print *,"freeSourceList"

  currentSource => sourceList
  downSourceList: do

     if(associated(currentSource%next)) then
        currentSource => currentSource%next
     else
        exit downSourceList
     end if

  end do downSourceList

  upSourceList: do 

     currentPixel => currentSource%pixelList
     downPixelList: do
        if(associated(currentPixel%next)) then
           currentPixel => currentPixel%next
        else
           exit downPixelList
        end if
     end do downPixelList
     
     upPixelList: do
        if(associated(currentPixel%previous)) then
            currentPixel =>  currentPixel%previous
            deallocate(currentPixel%next,stat = allocErr)
            call CheckDeallocation(allocErr, "currentpixel%next")
            nullify(currentPixel%next)
        else
            exit upPixelList
        end if
     end do upPixelList
     deallocate(currentPixel, stat = allocErr)
     call CheckDeallocation(allocErr, "currentPixel")
          
     nullify(currentPixel)

     if(associated(currentSource%previous)) then
         currentSource => currentSource%previous
         deallocate(currentSource%next, stat = allocErr)
         call CheckDeallocation(allocErr, "currentSource%next")
         nullify(currentSource%next)
     else
         deallocate(currentSource, stat = allocErr)
         call CheckDeallocation(allocErr, "currentSource")
         exit upSourceList
     end if

  end do upSourceList

  nullify(sourceList)

end subroutine freeSourceList
!******************************************
! subroutine getLinkedSources
!
! Routine to get nearest neighbour 
!******************************************
subroutine getLinkedSources(image, nx, ny, nSigma, minSigma, sourceList, nSourceList)

    integer, intent(in) :: nx, ny
    real, intent(inout) :: image(0 : nx - 1, 0 : ny - 1)
    real, intent(in) :: nSigma, minSigma
    type(src), pointer :: sourceList
    integer, intent(inout) :: nSourceList

    type(src), pointer :: currentSource
    type(srcpixel), pointer :: currentPixel !, nextPixel
 
    integer :: maxpos(2)
    integer :: xpos, ypos
    integer :: allocErr

    real :: tmp

!  print *,"getLinkedSources"

    maxpos = MAXLOC(image)

    xpos = maxpos(1) - 1
    ypos = maxpos(2) - 1

    ! If no pixels satisfy criterion

    if(image(xpos, ypos) < nSigma) then
        return
    endif

    tmp = image(xpos, ypos)

    ! We have some pixels that satisfy criterion

    nSourceList = 1
    allocate(sourceList, stat = allocErr)
    call CheckAllocation(allocErr, "sourceList")

    sourceList%numPixels = 0
    nullify(sourceList%previous)
    currentSource => sourceList

    !************************************************
    ! Loop round discrete sources to get final list
    !************************************************
    linkSources: do

        currentSource%sourceNum = nSourceList
        allocate(currentSource%pixelList, stat = allocErr)
        call CheckAllocation(allocErr, "pixelList (getLinkedSources)")

        ! Set the number of pixels to 1
     
        currentSource%numPixels = 1
     
        ! Now add in pixel numbers
     
        currentSource%pixelList%x = xpos
        currentSource%pixelList%y = ypos
     
        nullify(currentSource%pixelList%next)
        nullify(currentSource%next)
     
        image(xpos, ypos) = 0  ! Set to zero so next MAXLOC doesn't 
                         ! find this one

        ! Now find nearest associated pixels and add to pixel list

        ! Look in 8 directions eight check is recursive
        ! groups all associated pixels together
        
        currentPixel => currentSource%pixelList
        nullify(currentPixel%previous)
        call nearestNeighbourNew(image, nx, ny, xpos, ypos, minSigma, currentPixel)
              
        ! Next maximum datapoint

        maxpos = MAXLOC(image)     
        maxpos(1) = maxpos(1) - 1
        maxpos(2) = maxpos(2) - 1
        if(image(maxpos(1), maxpos(2)) .lt. nSigma) exit linkSources
        
        tmp = image(maxpos(1), maxpos(2))

        xpos = maxpos(1)
        ypos = maxpos(2)

        ! Now move onto next source
          
        nSourceList = nSourceList + 1
        allocate(currentSource%next,stat=allocErr)
        call CheckAllocation(allocErr, "pixelList (getLinkedSources)")

        currentSource%next%previous => currentSource
     
        currentSource => currentSource%next

    end do linkSources     

end subroutine getLinkedSources
   
!*************************************************
! subroutine nearestNeighbourNew
! Check nearest pixels without recursive call
!
!*************************************************
subroutine nearestNeighbourNew(image, nx, ny, xpos, ypos, nSigma, outPixel)

    integer, intent(in) :: nx,ny
    real, intent(inout) :: image(0 : nx - 1, 0 : ny - 1)
    integer, intent(in) :: xpos, ypos
    real, intent(in) :: nSigma
    type(srcpixel), pointer :: outPixel, nextPixel

    integer, allocatable :: there(:, :)
    type(srcPixel), pointer :: checkPixel
    integer :: allocErr
    integer :: k, l
    integer :: minXpos, minYpos, maxXpos, maxYpos

!  print *,"nearestNeighbourNew"

    allocate(there(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "there") 

    there=0
    !********************
    ! Do middle line
    !********************

    nextPixel => outPixel
    nextPixel%x = xpos
    nextPixel%y = ypos
    nullify(nextPixel%next)
    checkPixel => nextPixel

    loop: do

        call getMinMaxPosition(checkPixel%x, checkPixel%y, nx, ny, &
                               minXpos, maxXpos, minYpos, maxYpos)
     
        do k = minYpos, maxYpos
            do l = minXpos, maxXpos
                if(k .ne. checkPixel%x .and. l .ne. checkPixel%y) then
                    if(there(l, k) .eq. 0 .and. image(l, k) .gt. nSigma) then
                        there(l, k) = 1
                 
                        allocate(nextPixel%next, stat = allocErr)
                        call CheckAllocation(allocErr, "nextPixel%next (nearestNeighbourNew)")
                        nextPixel%next%previous => nextPixel
                        nextPixel => nextPixel%next
                        nextPixel%x = l
                        nextPixel%y = k
                        nullify(nextPixel%next)
                    end if
                end if
            end do
        end do

        !********************************************
        ! Move check position - if come to end exit
        !********************************************
        if(associated(checkPixel%next)) then
            checkPixel => checkPixel%next
        else
            exit loop
        end if

    end do loop

    !*******************************
    ! Removed used pixels from image
    !*******************************
    where(there .eq. 1) image = 0
  

    if(allocated(there)) then
        deallocate(there, stat = allocErr)
        call CheckDeallocation(allocErr, "there")
    end if
end subroutine nearestNeighbourNew

!*********************************************
! subroutine nearestNeighbour
! Check 8 sides for nearest pixels
!*********************************************
recursive subroutine nearestNeighbour(image, nx, ny, xpos, ypos, nSigma, &
     currentSource, currentPixel)

  integer, intent(in) :: nx,ny
  real, intent(inout) :: image(0 : nx - 1, 0 : ny - 1)
  integer, intent(in) :: xpos, ypos
  real, intent(in) :: nSigma
  type(src), pointer :: currentSource
  type(srcpixel), pointer :: currentPixel, nextPixel

  integer :: minXpos, maxXpos, minYpos, maxYpos
  integer :: allocErr

  integer :: i, j

!  print *,"NearestBNeighbour"

  ! Make sure that we don't violate arrays
  ! This should be faster the MIN/MAX?

  call getMinMaxPosition(xpos, ypos, nx, ny, minXpos, maxXpos, minYpos, maxYpos)

  ! Find next significant pixel

  do i = minYpos, maxYpos
     do j = minXpos, maxXpos
        if(i .ne. xpos .and. j .ne. ypos) then
           if(image(j, i) .ge. nSigma) then

              ! This is asociated with source

              currentSource%numPixels = currentSource%numPixels + 1
              allocate(currentPixel%next,stat=allocErr)
              call CheckAllocation(allocErr, "pixel%next (nearestNeighbour)")
              
              nextPixel => currentPixel%next
              nextPixel%x = j
              nextPixel%y = i
              image(j, i) = 0.  ! reset image so ignore this pixel
              nullify(nextPixel%next)

              ! Now go down a level (recursive)

              currentPixel => nextPixel
              call nearestNeighbour(image, nx, ny, j, i, nSigma, currentSource, &
                   currentPixel)

           end if
        end if
     end do
  end do

end subroutine nearestNeighbour

!*************************************************************
! subroutine nearestNeighbourImage
! Check 8 sides for nearest pixels
! In this case add data to level image matching source number
!*************************************************************
recursive subroutine nearestNeighbourImage(image, nx, ny, levelImage, &
     minXpos, maxXpos, minYpos, maxYpos, xpos, ypos, nSigma, sourceNumber)
 
  integer, intent(in) :: nx, ny
  real, intent(in) :: image(0 : nx - 1, 0 : ny - 1)
  integer(kind=int16), intent(inout) :: levelImage(0 : nx - 1, 0 : ny - 1)
  integer, intent(in) :: minXpos, maxXpos, minYpos, maxYpos
  integer, intent(in) :: xpos, ypos
  real, intent(in) :: nSigma
  integer, intent(in) :: sourceNumber

  integer :: minX, maxX, minY, maxY
  integer :: i, j

!  print *,"NearestNeighbourImage"

  ! Ensure we havn't strayed from allowed bounds

  if(xpos .lt. minXpos .or. xpos .gt. maxXpos) return
  if(ypos .lt. minYpos .or. ypos .gt. maxYpos) return

  ! Make sure that we don't violate arrays
  ! This should be faster the MIN/MAX?

  call getMinMaxPosition(xpos, ypos, nx, ny, minX, maxX, minY, maxY)

  ! Find next significant pixel

  do i = minYpos, maxYpos
     do j = minXpos, maxXpos
        if(i .ne. xpos .and. j .ne. ypos) then
           if(image(j, i) .ge. nSigma .and. levelImage(j, i) .eq. 0) then

              levelImage(j, i) = sourceNumber
              
              ! This is asociated with source

              call nearestNeighbourImage(image, nx, ny, levelImage, minXpos, &
                   maxXpos, minYpos, maxYpos, j, i, nSigma, sourceNumber)

           end if
        end if
     end do
  end do

end subroutine nearestNeighbourImage
!***********************************************
! subroutine getMinMaxPosition
!***********************************************
subroutine getMinMaxPosition(xpos,ypos,nx,ny,minXpos,maxXpos,minYpos,maxYpos)

    integer, intent(in) :: xpos, ypos
    integer, intent(in) :: nx, ny
    integer, intent(out) :: minXpos, maxXpos
    integer, intent(out) :: minYpos, maxYpos

!  print *,"getMinMaxPosition"

    if(xpos .gt. 0 .and. xpos .lt. nx-1) then
        minXpos = xpos-1
        maxXpos = xpos+1
    else if(xpos .eq. 0) then
        minXpos = 0
        maxXpos = 1
    else
        minXpos = nx-2
        maxXpos = nx-1
    end if

    if(ypos .gt. 0 .and. ypos .lt. ny - 1) then
        minYpos = ypos - 1
        maxYpos = ypos + 1
    else if (ypos .eq. 0) then
        minYpos = 0
        maxYpos = 1
    else
        minYpos = ny-2
        maxYpos = ny-1
    end if

end subroutine getMinMaxPosition

!********************************************************
! subroutine smoothIMage
! Routine to smooth image for background determination
!********************************************************
subroutine smoothImage(image, nx, ny, background, smoothSize)
 
    integer, intent(in) :: nx, ny
    real, intent(in)    :: image(0 : nx-1, 0 : ny - 1)
    real, intent(inout) :: background(0 : nx - 1, 0 : ny - 1)
    real, intent(in)    :: smoothSize

! internal variables

    integer             :: height
    integer             :: i

! For top hat
  
    real, allocatable   :: scan(:)
    integer, allocatable :: nscan(:)

    integer             :: allocErr

!  print *,"SmoothImage"

! for gaussian smooth 

    height = smoothSize
     
  ! Get scan array
  
    allocate(scan(-height : height), stat = allocErr)
    call CheckAllocation(allocErr, "scan")

    allocate(nscan(-height : height), stat = allocErr)
    call CheckAllocation(allocErr, "nscan")

  ! smoothTopHat deals with one X position at a time

    do i = 0, nx - 1
        call MovingBox(image, nx, ny, background, height, height, i, scan, nscan, 1)
    end do
  
  ! deallocate memory
    if(allocated(scan)) then
        deallocate(scan, stat = allocErr)
        call CheckDeallocation(allocErr, "scan")
    end if

    if(allocated(nscan)) then
        deallocate(nscan, stat = allocErr)
        call CheckDeallocation(allocErr, "nscan")
    end if

end subroutine smoothImage

!******************************************************************************
! subroutine MovingBox
! Top hat smoothing routine - does a strip in Y direction (best array access)
!******************************************************************************
subroutine MovingBox(image, nx, ny, backgroundTot, width, height, startX, scan, &
                     nscan,type)

    integer, intent(in) :: nx, ny
    real, intent(in)    :: image(0 : nx - 1, 0 : ny - 1)
    real, intent(inout) :: backgroundTot(0 : nx - 1, 0 : ny - 1)

    integer, intent(in) :: width, height     ! Size of smoothing area
    integer, intent(in) :: startX           ! starting point
    real, intent(inout) :: scan(-height : height)
    integer, intent(inout) :: nscan(-height : height)
    integer, intent(in) :: type
    !***********************
    ! internal variables
    ! for top hat smooth
    !***********************
    integer             :: i, j ,k
    real                :: running_total
    integer             :: nrunning_total
    integer             :: minXpos, maxXpos
    integer             :: minYpos, maxYpos
    integer             :: clearStart, clearEnd
    real                :: scanval

!  print *,"MovingBox"

    !************************************
    ! initialise arrays for this strip
    !************************************
    scan = 0
    nscan = 0

    ! X min and max's are fixed
     
    minXpos = MAX(startX - width, 0)
    maxXpos = MIN(startX + width, nx -1)

 
    !**************************************************************
    ! Get start and end of clear (ie. don't need MIN/MAX checks )
    !**************************************************************
    clearStart = MIN(height + 1, ny)

    if(height .gt. ny) then
        clearEnd = -1
    else
        clearEnd = ny - height - 1
    end if

    !******************************************
    ! At start need to build up running total  
    ! First one
    !******************************************
    minYpos = 0
    maxYpos = MIN(height, ny - 1)


    running_total= 0.0
    nrunning_total= 0

    forall (j = minYpos : maxYpos)
     
        scanval = 0.0
        forall (k = minXpos : maxXpos) scanval = scanval + image(k, j)
        
        scan(j) = scanval
        nscan(j) = maxXpos - minXpos + 1
        running_total = running_total + scanval
        nrunning_total = nrunning_total + nscan(j)
!     print *,running_total,nrunning_total,running_total/nrunning_total
     
    end forall
  
    if(type .eq. 1) then
        backgroundTot(startX, 0) = running_total / nrunning_total
    else
        backgroundTot(startX, 0) = running_total
    end if

  
    do i = 1, clearStart - 1
     
        ! Shift down arrays
     
        forall (j = minYpos : maxYpos)
            scan(j - 1) = scan(j)
            nscan(j - 1) = nscan(j)
        end forall
     
        ! reset min/max values
     
        minYpos = MAX(i - height, 0)
        maxYpos = MIN(i + height, ny - 1)

        !***********************
        ! Add in new entry
        !***********************     
        scanval = 0.0
        forall (k = minXpos : maxXpos) scanval = scanval + image(k, maxYpos)
        
        minYpos = minYpos - i
        maxYpos = maxYpos - i
        scan(maxYpos) = scanval
        nscan(maxYpos) = maxXpos - minXpos + 1
        running_total = running_total + scanval
        nrunning_total = nrunning_total + nscan(maxYpos)
     
        if(type .eq. 1) then
            backgroundTot(startX, i) = running_total / nrunning_total
        else
            backgroundTot(startX, i) = running_total
        end if
!     print *,running_total,nrunning_total,running_total/nrunning_total

    end do
    !********************************
    ! Now do the rest (not MIN/MAX)
    !********************************
    if(clearEnd .eq. -1) return
  

    ! Make sure we can hold whole smoothing area within data

    if(clearStart .lt. clearEnd) then
        do i = clearStart, clearEnd
     
            ! Shift down arrays and remove bottom line

            running_total = running_total - scan(-height)
            nrunning_total = nrunning_total - nscan(-height)
            forall (j = -height + 1 : height)
                scan(j - 1) = scan(j)
                nscan(j - 1) = nscan(j)
            end forall
            !*******************
            ! Add in new entry
            !******************
        
            maxYpos = i + height
            scanval = 0.0
            forall (k = minXpos : maxXpos) scanval = scanval + image(k, maxYpos)
            
            scan(height) = scanval
            nscan(height) = maxXpos - minXpos + 1
            running_total = running_total + scanval
            nrunning_total = nrunning_total + nscan(height)
        !     print *,running_total,nrunning_total,running_total/nrunning_total
        
            if(type .eq. 1) then
                backgroundTot(startX, i) = running_total / nrunning_total
            else
                backgroundTot(startX, i) = running_total
            end if

        end do
    end if

    !**********************
    ! And now do end part
    !**********************
  
    do i = clearEnd + 1, ny - 1
        running_total = running_total - scan(height - i + clearEnd + 1)
        nrunning_total = nrunning_total - nscan(height - i + clearEnd + 1)

!     print *,running_total,nrunning_total,running_total/nrunning_total
        if(type .eq. 1) then
            backgroundTot(startX, i) = running_total / nrunning_total
        else
            backgroundTot(startX, i) = running_total
        end if
    end do

end subroutine MovingBox

!**************************************************************
! function computeMedian
! computes the median of an image over the specific limits
!**************************************************************

function computeMedian(image, nx, ny, i1, i2, j1, j2) result (median)

  
    integer :: nx, ny
    real :: image(0 : nx - 1, 0 : ny - 1)
    integer :: i1, i2, j1, j2
    integer :: i, j
    real, allocatable :: tempArray(:)
    integer :: npts
    real :: median

    integer             :: allocErr

!  print *,"ComputeMedian"
   
    npts = (i2 - i1 + 1) * (j2 - j1 + 1)
    allocate(tempArray(npts), stat = allocErr)
    call CheckAllocation(allocErr, "tempArray")

    tempArray = 0
    npts = 0

    do i = i1, i2
        do j = j1, j2
            if(image(i, j) .ne. 0.0) then
                npts = npts + 1
                tempArray(npts) = image(i, j)
            end if
        end do
    end do

!    if(npts .ne. 0) then
!        call sort(npts, tempArray)
      
!        if (nint(real(npts) / 2.0) .ne. int(real(npts) / 2.0)) then
!            median = tempArray(nint(real(npts) / 2.0))
!        else
!            median = (tempArray(nint(npts / 2.0)) + &
!                      tempArray(nint(npts / 2.0) + 1)) / 2.0
!        end if
!    else
!        median = 1
!    end if

    median = computeMedianList(tempArray, npts)

    deallocate(tempArray, stat = allocErr)
    call CheckDeallocation(allocErr, "tempArray")
  
end function computeMedian

!********************************************************
! computeMedianList
! Get median of list only
! computes the median of an image over the specific limits
!********************************************************
function computeMedianList(list, npts) result (median)
  
    integer :: npts
    real :: list(npts)
    real :: median

    if(npts .ne. 0) then
        call sort(npts, list)
      

        if (npts .eq. (npts / 2) * 2) then   ! npts is even
            median = (list(npts / 2) + &
                      list(npts / 2 + 1)) / 2.0
        else   ! npts is odd
            median = list((npts + 1) / 2)
        end if
  
    else
        median = 1.0
    end if

end function computeMedianList

!*******************************
! computeMode
!******************************
subroutine computeMode(image, nx, ny, nSigma, mode)

    integer, intent(in) :: nx, ny
    real, intent(inout)    :: image(0 : nx - 1, 0 : ny - 1)
    real, intent(in)    :: nSigma
    real, intent(out)   :: mode
!  integer(kind=int32), intent(in) :: missingPixelValue

    integer :: i, j, k

    real, allocatable :: list(:), oldList(:)
    integer :: nlist
    integer :: allocStat
    integer :: m

    real :: sigma
    real :: median, mean
    real :: variance, originalSigma, testSigma
    real :: lowValue, highValue
    real :: imageVal
    integer :: intImageVal
    integer :: nhist
    integer, allocatable :: histogram(:)
    integer :: maxHistPos(1)
    logical :: allocHist
    integer :: allocErr

    allocate(list(nx * ny), stat = allocStat)
    call CheckAllocation(allocStat, "list")

    allocate(oldList(nx * ny), stat = allocStat)
    call CheckAllocation(allocStat, "oldlist")

    nhist = MAXVAL(image)

    allocHist = .false.
    if(nhist .gt. 30) then
        allocate(histogram(0 : nhist + 1), stat = allocStat)
        call CheckAllocation(allocStat, "histogram")
 
        histogram = 0
        allocHist = .true.
    end if
 
    nlist = 0
    mean = 0.0

    do i = 0, ny - 1
        do j= 0, nx - 1

            nlist = nlist + 1
            imageVal = image(j, i)
            list(nlist) = image(j, i) 
            mean = mean + image(j, i)

            if(allocHist) then
                intImageVal = int(imageVal)
                histogram(intImageVal) = histogram(intImageVal) + 1
            end if
        end do
    end do 
 
    mean = mean / nlist

    if(allocHist) then
        maxHistPos = MAXLOC(histogram)

        mode = maxHistPos(1)
    else
        mode = mean / nlist
    end if  
 
    !**************************************
    ! Get variance and std deviation
    !**************************************
    variance = 0.0
    forall (k = 1 : nlist) variance = variance + (list(k) - mode) ** 2
    
    variance = variance / (nlist - 1)
 
    originalSigma = sqrt(variance)
    lowValue = mode - nSigma * originalSigma
    highValue = mode + nSIgma * originalSigma

    !**********************************************
    ! reject data at more than nSigma from data
    !**********************************************
    testSigma = originalSigma
    sigmaClip: do 

        oldlist = list
        m = 0
        mean = 0.
        do k = 1, nlist
            if(oldList(k) .ge. lowValue .and. oldList(k) .le. highValue) then
                m = m + 1
                list(m) = oldList(k)
                mean = mean + list(m)
            end if
        end do     
 
        nlist = m
        if(nlist .lt. 20) exit sigmaClip
       
        mean = mean / nlist
        nhist = MAXVAL(list)

        if(nhist .gt. 30 .and. allocHist) then
            histogram = 0
            forall  (k= 1 : nlist)
                intImageVal = int(list(k))

                histogram(intImageVal) = histogram(intImageVal) + 1
            end forall
            maxHistPos = MAXLOC(histogram)
            mode = maxHistPos(1)        
        else
            mode = mean
        end if
 
        !*******************************
        ! Now do sigma clipping
        ! Get variance and std deviation
        !********************************
        variance = 0.0
        forall (k = 1 : nlist) variance = variance + (list(k) - mode) ** 2
        
        variance = variance / (nlist-1)
 
        sigma = sqrt(variance)
        lowValue = mode - nSigma * sigma
        highValue = mode + nSigma * sigma
        if(abs((sigma - testSigma) / sigma) .lt. 0.2) exit sigmaClip
        
        testSigma = sigma

    end do sigmaClip
 
    if(abs((sigma - originalSigma) / sigma) .gt. 0.2) then
        median = computeMedianList(list, nlist)
        mode = 2.5 * median - 1.5 * mean
    else
        mode = mean
    end if
 

    if(allocHist) then
        if(allocated(histogram)) then
            deallocate(histogram, stat = allocErr)
            call CheckDeallocation(allocErr, "histogram")
        end if
    end if

    deallocate(list, stat = allocErr)
    call CheckDeallocation(allocErr, "list")
  
    deallocate(oldList, stat = allocErr)
    call CheckDeallocation(allocErr, "oldList")
   

end subroutine computeMode
!***************************************************
! computeBackground
! Computes the background using a median box filter
!***************************************************
subroutine computeBackground(image, nx, ny, background, smBackground, &
     medianBackground, nSigma)

    integer, intent(in) :: nx, ny
    real, intent(in)    :: image(0 : nx - 1, 0 : ny - 1)
    real, intent(inout) :: background(0 : nx - 1, 0 : ny -1)
    real, intent(in)    :: smBackground(0 : nx - 1, 0 : ny - 1)
    real, intent(out)   :: medianBackground
    real, intent(in)    :: nSigma
    real                :: sigmaBackground
    integer, parameter  :: nBox = 5
    real, allocatable   :: tmpBackground(:, :)
    logical, allocatable :: mask(:, :)
    real                 :: rmsval, newrmsval
    integer             :: allocErr

!  print *,"ComputeBackground"

    allocate(tmpBackground(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "tmpBackground")

    allocate(mask(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "mask")

    tmpBackground = smBackground


    sigmaBackground = sqrt(medianBackground)

    !**********************************************************************
    ! Get rid of bright sources on scale of smoothing - use sigma clipping
    !
    ! sigma clip until got down to a reasoanble value
    !*********************************************************************
    rmsval = sqrt(SUM((smBackground - medianBackground) ** 2) / size(smBackground))
    sigma_clip: do 
        
        where ( (smBackground - medianBackground) < nSigma * rmsval) 
            mask = .true.
        elsewhere
            mask = .false.
        end where

        newrmsval = rmsval
        if(count(mask) .gt. 1) then
            rmsval = sqrt(SUM((smBackground - medianBackground) ** 2, mask) / &
                         count(mask))
        else
            exit sigma_clip
        end if
        if((newrmsval - rmsval) / newrmsval .lt. 0.5) exit sigma_clip
        
    end do sigma_clip

    !****************************************
    ! Now put median value where sources are
    !****************************************
    where ( (smBackground - medianBackground) .gt. nSigma * rmsval) tmpBackground = medianBackground
    
    !*********************************************************
    ! Get rid of bright sources smaller than smoothing scale
    !*********************************************************
    where ( ((image - medianBackground) / sigmaBackground) .gt. nSigma) tmpBackground = medianBackground
    
    background = tmpBackground

    if(allocated(tmpBackground)) then
        deallocate(tmpBackground, stat = allocErr)
        call CheckDeallocation(allocErr, "tmpBackground")
    end if

    if(allocated(mask)) then
        deallocate(mask, stat = allocErr)
        call CheckDeallocation(allocErr, "mask")
    end if

end subroutine computeBackground
!*********************************************
!***************************************************
! AddReal32Cololumn
! Adds a column of real32 data to a fits table.
!***************************************************
subroutine AddReal32Column(tableSet, heading, comment, &
        size, array, nullValue)
    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: heading, comment
    integer, intent(in) :: size
    real(kind=single), intent(in), dimension(:) :: array
    real(kind=single), intent(in) :: nullValue
    real(kind=single), pointer :: real32Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)
 
    pcolumn = addColumn(tableSet, heading, real32, comment, ndims)
    real32Col => real32Data(pcolumn)

    do i = 1, size
        if (array(i) .eq. nullValue) then
            call setToNull(pcolumn, i - 1)
        else
            real32Col(i - 1) = array(i)
        end if
  
    end do
    call release(pcolumn)

end subroutine AddReal32Column
  
!***************************************************
! AddReal64Column
! Adds a column of real64 data to a fits  table.
!***************************************************
subroutine AddReal64Column(tableSet, heading, comment, &
        size, array, nullValue)
    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: heading, comment
    integer, intent(in) :: size

    real(kind=double), intent(in), dimension(1 : size) :: array
    real(kind=single), intent(in) :: nullValue
    real(kind=double), pointer :: real64Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)

    pcolumn = addColumn(tableSet, heading, real64, comment, ndims)
    real64Col => real64Data(pcolumn)

    do i = 1, size
        if (array(i) .eq. nullValue) then
            call setToNull(pcolumn, i - 1)
        else
            real64Col(i - 1) = array(i)
        end if
  
    end do
    call release(pcolumn)

end subroutine AddReal64Column
!***************************************************
! AddInt8Column
! Adds a column of integer8 data to a fits table.
!***************************************************
subroutine AddInt8Column(tableSet, heading, comment, &
        size, array)
    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: heading, comment
    integer, intent(in) :: size
    integer(kind=int8), intent(in), dimension(:) :: array
    integer(kind=int8), pointer :: int8Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)

    pcolumn = addColumn(tableSet, heading, integer8, comment, ndims)
    int8Col => int8Data(pcolumn)

    forall (i = 1 : size) int8Col(i - 1) = array(i)
    

    call release(pcolumn)

end subroutine AddInt8Column
end module omdetect_mod









