!******************************************************************
! Module omdetectnumerics_mod
!
! Contains various subroutines for numerical procedures.
!******************************************************************
module omdetectnumerics_mod
  use errorHandling
  use omdetectconstants_mod
contains
!********************************************************************
! Subroutine fitGaussian
! Fitting routine to get source parameters to assign pixel parameters
!********************************************************************
subroutine fitGaussian(image, nx, ny, levelImage, i1, i2, j1, j2, params)

    integer, intent(in) :: nx, ny
    real, intent(in) :: image(nx, ny)
    integer, intent(in) :: i1, i2, j1, j2
    integer, intent(in) :: levelImage(i1 : i2, j1 : j2)
    real, intent(out) :: params(6)

    real :: fwhm_x, fwhm_y
    integer :: maxvalue(2)
    real :: pmat(7, 6)
    real :: ptest(6)
    real :: yvec(7)
    real :: frac(6)
    integer :: m,n

    integer :: iter
    logical :: ok

  ! Set up frac array

    frac(1) = 1.05       ! fg
    frac(2) = 1.05       ! sig_x
    frac(3) = 1.05       ! sig_y
    frac(4) = 1.01       !pa
    frac(5) = 1.0005     ! x
    frac(6) = 1.0005     ! y
  
  ! set up initial guesses for gaussian

    fwhm_x = real(i2 - i1) / 1.5
    fwhm_y = real(j2 - j1) / 1.5
    fwhm_x = max(fwhm_x, 0.5)
    fwhm_y = max(fwhm_y, 0.5)
  
    params(1) = SUM(image(i1 : i2, j1 : j2))
    params(2) = fwhm_x / 2.35
    params(3) = fwhm_y / 2.35
    params(4) = 0.0
    maxvalue = MAXLOC(image(i1 : i2, j1 : j2))
    params(5) = real(maxvalue(1)) + i1
    params(6) = real(maxvalue(2)) + j1

  ! The following code defines the nodes of the simplex
  
    pmat(1,1) = params(1)                        
    pmat(1,2) = params(2)
    pmat(1,3) = params(3)
    pmat(1,4) = params(4)
    pmat(1,5) = params(5)
    pmat(1,6) = params(6)
  
    ! define nodes of simplex
  
    do m = 2 , 7
!        do n = 1 , 6
        forall (n = 1: 6) pmat(m, n) = params(n)
!        end do
        pmat(m, m - 1) = pmat(m, m - 1) * frac(m - 1)
    end do
  
    do m = 1 , 7
!        do n = 1 , 6
        forall (n = 1 : 6) ptest(n) = pmat(m, n)
!        end do
        yvec(m) = gauss2d(ptest, image, nx, ny, levelImage, i1, i2, j1, j2)
    end do

  ! Fit the gaussian using the downhill simplex method
  
  call simplex(pmat, yvec, 7, 6, 6, 1.e-4, gauss2d, iter, image, nx, ny, &
       levelImage, i1, i2, j1, j2, ok)
  
  ! The solution is taken as the mean of the simplex nodes
  
  params = 0.
  do m = 1 , 7
!     do n = 1 , 6
      forall (n = 1 : 6) params(n) = params(n) + pmat(m, n)
!     enddo
  end do
  params = params  / 7.
  
end subroutine fitGaussian
!************************************************
! fitGaussianMoments
! Fits a two-dimensional gaussian to the sourfce counts
!************************************************
subroutine fitGaussianMoments(image, nx, ny, levelImage, i1, i2, j1, j2, params)

    integer, intent(in) :: nx, ny
    real, intent(in) :: image(nx, ny)
    integer, intent(in) :: i1, i2, j1, j2
    integer, intent(in) :: levelImage(i1 : i2, j1 : j2)
    real, intent(out) :: params(6)

    real :: centX, centY
    real :: x2, y2, xy
    real :: total, total2
    real :: imageVal
    real :: val1, val2
    real :: aValue, bValue
    real :: theta  !, theta1, theta2
    
    integer :: i, j

    ! Don't fit - use moments to see if it works better.

    centX = 0
    centY = 0
    x2 = 0
    y2 = 0
    xy = 0.
    total = 0
    total2 = 0.
    do i = j1 , j2
        do j = i1 , i2
            if(levelimage(j, i) .gt. 0) then
                imageVal = image(j, i)
                total = total + imageVal
                total2 = total2 + imageVal * imageVal
                centX = centX + imageVal * j
                centY = centY + imageVal * j
                x2 = x2 + imageVal * j * j
                y2 = y2 + imageVal * i * i
            end if
        end do
    end do

    centX = centX / total
    centY = centY / total

    x2 = x2 / total - centX * centX
    y2 = y2 / total - centY * centY
    xy = xy / total - centX * centY

    val1 = (x2 + y2) / 2.
    val2 = sqrt(((x2 - y2) / 2.) ** 2 + xy ** 2)

    aValue = sqrt(val1 + val2)
    bValue = sqrt(val1 - val2)

  ! Sort out angle
    theta = Angle(x2, y2, xy)

    params(1) = total
    params(2) = aValue ! fwhm_x
    params(3) = bValue ! fwhm_b
    params(4) = theta  ! PA
    params(5) = centX  ! x
    params(6) = centY  ! y

end subroutine fitGaussianMoments
!********************************************************************
! fitGaussian2
!********************************************************************
subroutine fitGaussian2(image, nx, ny, nUseSource, inParams, xRange, yRange)

    integer, intent(in) :: nx,ny
    integer, intent(in) :: nUseSource
    real, intent(in) :: image(nx, ny, nUseSource)
    real, intent(inout) :: inParams(6, nUseSource)
    integer, intent(in) :: xRange(2, nUseSource), yRange(2, nUseSource)

    real, allocatable :: newImage(:, :)
    integer, allocatable :: levelImage(:, :)
    integer :: allocErr
    real :: gParams(6)
    integer :: i, j, k

    allocate(newImage(nx, ny), stat=allocErr)
    call CheckAllocation(allocErr, "newImage (fitGaussian2)")
  

    do i = 1, nUseSource
  
        newImage = image(1 : nx, 1 : ny, i)

        allocate(levelImage(xRange(1, i) : xrange(2, i),  & 
                         yRange(1, i) : yRange(2, i)), &
                         stat=allocErr)
        call CheckAllocation(allocErr, "levelImage(fitGaussian)")
    
     
        levelImage = 0
        do j = yRange(1, i), yRange(2, i)
            do k = xRange(1, i), xRange(2, i)
                if(newImage(k, j) .ne. pixelIgnoreFlag) then

                    levelImage(k, j) = 1
                else
                    levelImage(k, j) = pixelIgnoreFlag
                end if
            end do
        end do
        !**********************************************
        ! Fit 2-dimensional gaussian to source counts
        !**********************************************
        call fitGaussian(newImage, nx, ny, levelImage, xRange(1, i), xRange(2, i), &
                         yRange(1, i), yRange(2, i), gParams)
        deallocate(levelImage)

        inParams(1, i) = gParams(1)
        inParams(2, i) = gParams(2)
        inParams(3, i) = gParams(3)
        inParams(4, i) = gParams(4)
        inParams(5, i) = gParams(5)
        inParams(6, i) = gParams(6)

    end do

    deallocate(newImage)
  
end subroutine fitGaussian2
!***************************************************************
! simplex
! downhill simplex routine
!***************************************************************
subroutine simplex(pmatrix, yvec, msize, nsize, ndimen, simptol, exfunk, iter,&
                   osw, nx, ny, level, i1, i2, j1, j2 ,ok)


    logical :: ok

    integer,parameter :: nmax = 20
    integer,parameter :: itmax = 1000
    real,parameter    :: beta = 0.5
    real,parameter    :: alpha = 1.
    real,parameter    :: gamma = 2.0
  
    integer :: i1, i2, j1, j2
    integer :: level(i1 : i2, j1 : j2)

    integer :: msize
    integer :: nsize
    integer :: mpts
    integer :: ndimen
    integer :: iter
    integer :: ilo
    integer :: j

    integer  :: nx
    integer  :: ny
    real     :: osw(nx, ny)

    real :: pmatrix(msize, nsize)
    real :: yvec(msize)
    real :: pr(nmax)
    real :: prr(nmax)
    real :: pbar(nmax)

    integer :: ihi
    integer :: inhi
    integer :: i
    real    :: simptol
    real    :: rtol 
    real    :: ypr
    real    :: yprr

    real    :: exfunk
    external exfunk

    ok = .true.

    mpts = ndimen + 1
    iter = 0

    do while(.true.)
        ilo = 1

        if( yvec(1) .gt. yvec(2) ) then
            ihi = 1
            inhi = 2
        else
            ihi = 2
            inhi = 1
        end if

        do  i = 1 , mpts
            if (yvec(i) .lt. yvec(ilo)) ilo = i
            if (yvec(i) .gt. yvec(ihi)) then
                inhi = ihi
                ihi = i
            else if (yvec(i) .gt. yvec(inhi)) then
                if(i .ne. ihi) inhi = i
            end if
        end do

        rtol = 2.0 * abs(yvec(ihi) - yvec(ilo)) / (abs(yvec(ihi)) + abs(yvec(ilo)))

        if (rtol .lt. simptol) EXIT

        if (iter .eq. itmax) then
            ok = .false.
            EXIT
        end if
        iter = iter + 1

        pbar = 0.0

        do i = 1,  mpts 
            if (i .ne. ihi) then 
                do j = 1 , ndimen 
                    pbar(j) = pbar(j) + pmatrix(i, j)
                end do
            end if 
        end do           
  

        pbar = pbar / real(ndimen)

        forall (j = 1 : ndimen) pr(j) = (1. + alpha) * pbar(j) - alpha * pmatrix(ihi, j)
     

        ypr = exfunk(pr, osw, nx, ny, level, i1 ,i2, j1 ,j2)

        if (ypr .le. yvec(ilo)) then

            forall  (j = 1 : ndimen) prr(j) = gamma *pr(j) + (1. - gamma) * pbar(j)
       

            yprr = exfunk(prr, osw, nx, ny, level, i1, i2, j1, j2)

            if (yprr .lt. yvec(ilo)) then
                forall (j = 1 : ndimen) pmatrix(ihi, j) = prr(j)
           
                yvec(ihi) = yprr
        else
           forall (j = 1 : ndimen) pmatrix(ihi, j) = pr(j)
           
           yvec(ihi) = ypr
        end if

     else if (ypr .ge. yvec(inhi)) then

        if (ypr .lt. yvec(ihi)) then
           forall (j = 1 : ndimen) pmatrix(ihi, j) = pr(j)
           
           yvec(ihi) =  ypr
        end if

        forall (j=1 : ndimen) prr(j) = beta * pmatrix(ihi, j) + (1. - beta) * pbar(j)
       

        yprr = exfunk(prr, osw, nx, ny, level, i1, i2, j1, j2)

        if (yprr .lt. yvec(ihi)) then
           forall (j = 1 : ndimen) pmatrix(ihi, j) = prr(j)
           
           yvec(ihi) = yprr
        else
           do i = 1, mpts
              if(i .ne. ilo)then
                 forall (j = 1 : ndimen)
                    pr(j) = 0.5 * (pmatrix(i, j) + pmatrix(ilo, j))
                    pmatrix(i, j) = pr(j)
                 end forall
                 yvec(i) = exfunk(pr, osw, nx, ny, level, i1, i2, j1, j2)
              end if
           end do
        end if

     else

        forall (j = 1 : ndimen) pmatrix(ihi, j) = pr(j)
        

        yvec(ihi) = ypr

     end if

  end do

end subroutine simplex
!*******************************************************
! gauss2d
!*******************************************************
real function gauss2d(param, osw, nx, ny, level, i1, i2, j1, j2)

    implicit none

    real param(*)                  ! the Gaussians parameters
    integer nx, ny                  ! the osw image size
    real osw(nx, ny)                ! the osw data array
    integer i, j, n                  ! counters
    real flux                      ! the source flux
    real x, y                       ! the x,y position of a pixel
    real x0, y0                     ! the x,y position of the gaussian
    real sig_x, sig_y              ! the sigma of the gaussians
    real pa
    real tot                       ! total
    real f                         ! flux
    real ex1, ex2, xDash, yDash, delta, r
    integer :: i1, i2, j1, j2
    integer :: level(i1 : i2, j1 : j2)
    real :: oswval
  ! set up the Gaussian parameters
    flux = param(1)
    sig_x = param(2)
    sig_y = param(3)
    pa = param(4)
    x0 = param(5)
    y0 = param(6)
    tot = 0.
    n = 0

    if (sig_x * sig_y .ne. 0.0) then
        do i = j1 , j2
            y = real(i) + 0.5
           
            do j = i1 , i2
                if(level(j, i) .gt. 0) then
                    x = real(j) + 0.5
                    r = sqrt((x - x0) ** 2 + (y - y0) ** 2)
                    delta = atan2(y - y0, x - x0)
              
                    xDash = r * cos(delta - pa)
                    yDash = r * sin(delta - pa)
              
                    ex1 = -((xDash / sig_x) ** 2)
              
                    ex2 = -((yDash / sig_y) ** 2)
                            
                    f =  0.159154943 / (sig_x * sig_y)
              
                    f = f * flux * exp(0.5 * (ex1 + ex2))
              
                    oswval = osw(j, i)
                    if(oswval .gt. 0) then
                        tot = tot + (f - oswval) ** 2 / oswval
                    else if(oswval .eq. 0) then
                        tot = tot + (f - oswval) ** 2
                    end if
              
                    n = n + 1
                end if
            end do
        end do
    else
        tot = 1.e20
    end if


    gauss2d  = tot
end function gauss2d

!********************************
! function Angle
! 
!********************************
real function Angle(x, y, xy)
    real, intent(in) :: x, y, xy
    real theta1, theta2 

    if(x .ne. y) then
        theta1 = atan(2 * xy / (x - y)) / 2.0
        theta2 = theta1 + HALF_PI
        if(theta2 .gt. HALF_PI) theta2 = theta2 - PI
        if(xy .ge. 0.0) then
            if(theta1 .ge. 0.0) then
                Angle = theta1
            else
                Angle = theta2
            end if
        else
            if(theta1 .lt. 0.0) then
                Angle = theta1
            else
                Angle = theta2
            end if
        end if
    else
         Angle = sign(real(HALF_PI), xy) / 2.0
    end if
end function Angle

!***********************************************
! function gammln
!***********************************************
real function gammln(xx) result(output)

    implicit none
    real, intent(in) :: xx

  ! internal variables

    integer j
    real(kind=double) :: ser, tmp, x, y
    real(kind=double), save :: cof(6), stp
  
!  print *,"gammln"

    data cof/76.18009172947149_double, -86.50532032941677_double, 24.01409824083091d0,&
        -1.231739572450155_double, 0.1208650973866179d-2, -0.5395239384953d-5/  
    data stp/2.5066282746310005_double/
  
    x = xx
    y = x
    tmp = x + 5.5_double
    tmp = (x + 0.5d0) * log(tmp) - tmp
    ser = 1.000000000190015_double
 
    do j = 1, 6
        y = y + 1.0_double
        ser = ser + cof(j) / y
    end do
    output = tmp + log(stp * ser / x)
  
end function gammln

!*********************************************
! function lnFactrl
! Log of n!
!*********************************************
real function lnFactrl(n) result(output)

    implicit none

    integer, intent(in)  :: n
  
  ! internal variables 
   
    real, save :: a(100)
    data a /100 * -1.0/

!   print *,"lnfractrl"
  
    if(n .lt. 0)then
        call fatal("lnFactrl","Incorrect input n < 0")
    else if(n .le. 99) then
        if(a(n + 1) .lt. 0) a(n + 1) = gammln(n + 1.0)
        output = a(n + 1)      
    else
        output = gammln(n + 1.0)
    end if

end function lnFactrl

! errf - a=0.5 is explicit

!***************************************************************
! function newerrf
! New error function routine that copes with n sighma better
!
!***************************************************************
double precision function newerrf(x) result(resultval)

    real, intent(in) :: x
    real(kind=double) :: double_x
    real(kind=double) :: z, t, expval
  
! Use erfc function to get probability

    double_x = x / sqrt(2.0_double)
    z = abs(double_x)
    t = 1.d0 / (1.d0 + 0.5_double * z)
    expval = -z * z - 1.26551223_double + t * (1.00002368_double + t * (0.37409196_double + t * (0.09678418_double + &
       t * (-0.18628806_double + t * (0.27886807_double + t * (-1.13520398_double + t * (1.48851587_double + &
       t * (-0.82215223_double + t * 0.17087277_double))))))))
    resultval = log(t) + expval
    resultval = exp(resultval)
    if(resultval .lt. 0.0_double) resultval = 2 - resultval

end function newerrf
!***************************************************
! function errfx
!
!***************************************************

double precision function errf(x) result(output)

    real :: x
    real(kind=double) :: gammcf, gamser, xx

!  print *,"errf"

    xx = DBLE(x)
    if(xx .lt. 0.0_double) then
        call fatal("gammp","Incorrect arguments")
    end if
    if(xx .lt. 1.5_double) then
        call lnGserErf(gamser, xx * xx)
        output = exp(gamser)
    else
        call lnGcfErf(gammcf, xx * xx)
        output = 1 - exp(gammcf)
    end if

end function Errf
!***************************************
! subroutine lnGserErf 
! gser routine with a=0.5 explicit
!***************************************
subroutine lnGserErf(gamser,x) 

    real(kind=double) :: gamser,x

    integer :: n
    real(kind=double) :: ap, del, sum, gln
    real(kind=double) :: output

!  gln = detectGammln(0.5)
!  make gln the value of gammln(0.5)

    data gln /0.572365_double/

!  print *,"gsererf"

    if(x .le. 0.0_double) then
        if(x .lt. 0.0_double) then
            call fatal("gser_erf", "x < 0 in gser")
        end if
        output = 0.0_double
        return
    end if
  
    gamser = -1d8
    ap = 0.5
    sum = 1. / ap
    del = sum
    erf_loop: do n = 1, 100
        ap =ap + 1.
        del = del * x / ap
        sum = sum + del
        if(abs(del) .lt. abs(sum) * 3.d-7) then
            gamser = log(sum) +(-x + 0.5d0 *log(x) - gln)
            exit erf_loop
        end if
    end do erf_loop

    if(gamser .eq. -1d8) call fatal("gser_erf","Number of iteration exceeded")
  

end subroutine lnGserErf
!************************************************
! subroutine lnGcfErf
! gcf routine with a=0.5 explicit
!************************************************
subroutine lnGcfErf(gammcf, x)

    real(kind=double) :: gammcf, x
    integer :: i
    real(kind=double) :: an, b, c, d, del, h, gln

!  gln = detectGammln(0.5)
!  make gln the value of gammln(0.5)

    data gln /0.572365_double/

!  print *,"gcferf"

    gammcf = -1d8

    b = x + 0.5_double
    c = 1d30
    d = 1. / b
    h = d
    gcf_loop: do i = 1, 100
        an = -i * (i - 0.5_double)
        b = b + 2
        d = an * d + b
        if(abs(d) .lt. 1d-30) d = 1d-30
     
        c = b + an / c
        if(abs(c) .lt. 1d-30) c = 1d-30
     
        d = 1.0 / d
        del = d * c
        h = h * del
        if(abs(del-1.) .lt. 3.d-7) then
        gammcf = (-x + 0.5d0 * log(x) - gln) + log(h)
     end if
  end do gcf_loop
  if(gammcf .eq. -1d8) then
     call fatal("gcf_erf","Number of interation exceeded")
  end if

end subroutine lnGcfErf
!*****************************************************************
! function getSigmaFromProb
! Routine to get sigma for a given (gaussian) probability
! This prob won;t work since errf can't get beyond 7 sigma or so...
!******************************************************************
real function getSigmaFromProb(prob, guess) result(output)

    real, intent(in) :: prob, guess
    real :: lowRange, highRange

!  print *,"getsigmafromprob"

  ! error function gives prob for given sigma so need to find correct sigma
  ! use failsafe method
  
    lowRange = guess * 0.1
    highRange = guess * 10

    output = rtsafe(lowRange, highRange, 1e-4, prob)

end function getSigmaFromProb

!*******************************************
! subroutine getFunction
! Returns the function (root) to get sigma
!*******************************************
subroutine getFunction(x1,f,df,prob)

    real, intent(in) :: x1, prob
    real, intent(out) :: f, df
    real :: inx

!  print *,"getfunction"

    f = Errf(x1)
    inx = x1 * 1.00001
    df = (Errf(inx) - f) / (0.00001 * x1)
    f = f + prob - 1

end subroutine getFunction
!***********************************
! function rtsafe
!***********************************
real function rtsafe(x1, x2, xacc, prob) result(output)

    real, intent(in) :: x1, x2, xacc, prob
    integer :: j
    real :: df, dx, dxold, f, fh, fl, temp, xh, xl

!  print *,"rtsafe"

    call getFunction(x1, fl, df, prob)
    call getFunction(x2, fh, df, prob)
  
    if((fl .gt. 0.0 .and. fh .gt. 0.0) .or. (fl .lt. 0.0 .and. fh .lt. 0.0)) then
        output = x2
        return
    else if(fh .eq. 0) then
        output = x2
        return
    else if(fl .lt. 0.0) then
        xl = x1
        xh = x2
    else
        xl = x2
        xh = x1
    end if

    output = 0.5 * (x1 + x2)
    dxold = abs(x2 - x1)
    dx = dxold
    call getFunction(output, f, df, prob)
    do j = 1, 100
        if(((output - xh) * df -f) * ((output - xl) * df - f) .ge. 0 .or. &
            abs(2.0 * f) .gt. abs(dxold * df)) then
            dxold =dx
            dx = 0.5 *(xh - xl)
            output = xl + dx
            if(xl .eq. output) return
        else
            dxold = dx
            dx = f / df
            temp = output
            output = output - dx
            if(temp .eq. output) return
        end if
        if(abs(dx) .lt. xacc) return
        call getFunction(output, f, df, prob)
        if(f .lt. 0.0) then
            xl = output
        else
            xh = output
        end if
    end do 

    call fatal("rtsafe","Too many iterations")

end function rtsafe

!***************************************************
! CheckAllocation
!***************************************************
subroutine CheckAllocation(status, string)
    integer, intent(in) ::  status
    character(len=*), intent(in) :: string
    character(len=100) :: comment
   
    if(status .ne. 0) then
       write(comment, '(a32, a20)') "Could not allocate memory for : ", string 
       write (errstr, *) comment
       call fatal("allocErr") 
    end if
    noAllocations = noAllocations + 1
end subroutine CheckAllocation
!***************************************************
! CheckDeallocation
!***************************************************
subroutine CheckDeallocation(status, string)
    integer, intent(in) ::  status
    character(len=*), intent(in) :: string
    character(len=100) :: comment
   
    if(status .ne. 0) then
       write(comment, '(a31, a20)') "Failed to release memory for : ", trim(string) 
       write (errstr, *) comment
       call warning("allocErr", comment) 
    end if
    noDeallocations = noDeallocations + 1
end subroutine CheckDeallocation 

end module omdetectnumerics_mod




