!*******************************************
! The second half of omdetect
!******************************************
module omdetectSourceParams_mod
   
    use dal
    use errorHandling
    use omdetectconstants_mod
    use omdetectnumerics_mod
    implicit none

    private
    public detectParams
    public sourceParams
    public CheckAllocation            ! Subroutine for checking that memory for arrays allocated okay.
    public CheckDeallocation          ! Subroutine for checking that memory for arrays deallocated okay. 
   
    type srcpixel                              ! Pixel list node.
        integer(kind=int32) x                  ! X position.
        integer(kind=int32) y                  ! Y position.
        type(srcpixel), pointer :: next        ! Next pixel in source.
        type(srcpixel), pointer :: previous    ! Previous one (mem removal)
    end type srcpixel
!*******************************************
! A structure to store various parameters
! fo each source detected in the image
!******************************************* 
    type sourceParams
       real :: x                ! X pixel coordinate
       real :: y                ! Y pixel coordinate
       real :: errX             ! Error in x
       real :: errY             ! Error in y
       real :: PosAngle         ! Position angle of major-axis of source ellipse
       real :: semiMajor        ! Semi-major axis of ellipse
       real :: semiMinor        ! Semi-minor axis of ellipse
       real :: countrate        ! Count-rate (background subtracted)
       real :: countrateErr     ! The count-rate error
       real :: backGroundRate   ! The background count-rate
       integer(kind=int32) :: noSourcePixels  ! The number of pixels the source includes
       real :: signif
       integer :: Cflag 
    end type sourceParams
!****************************************
! A structure for ?
!**************************************** 
    type treeStruct
        integer :: nSources   ! The number of sources
        integer :: level
        integer :: minX
        integer :: maxX
        integer :: minY
        integer :: maxY
        integer, pointer :: levelImage(:,:)
        type(treeStruct), pointer :: next
        type(treeStruct), pointer :: downLevel
        type(treeStruct), pointer :: upLevel
    end type treeStruct

    integer :: maxXsize = 7
    integer :: maxYsize = 7
 

contains
!************************************************************************
! Subroutine detectParams
! Purpose:
! Parameters:
! inImage - the input image
! nx, ny - the dimensions of the image, background and level input images.
! inBackground - the input background image.
! inLevelImage - the input image identifying the pixels connected with
! each of the sources first found (each pixel associated with a source has
! been assigned the source number value.
! scrNo - the input source number.
! xlow, xhigh, ylow, yhigh - the pixel boundaries that contain the source.
! nSigma -
! fraction -
! nsourcePars - the number of sources that the original source has been
! decomposed into.
! sourcePars - an array of structures containing the source parameters.
!*************************************************************************
subroutine detectParams(nx, ny, inImage, inBackground, inLevelImage, srcNo, &
     xlow, xhigh, ylow, yhigh, nSigma, fraction, nSourcePars, sourcePars)

    integer, intent(in) :: nx, ny
    real, intent(in) :: inImage(nx, ny)
    real, intent(in) :: inBackground(nx, ny)
    integer(kind=int16), intent(in) :: inLevelImage(nx, ny)
    integer, intent(in) :: srcNo
    integer, intent(in) :: xlow, xhigh, ylow, yhigh
    real, intent(in) :: nSigma
    real, intent(in) :: fraction
    type(sourceParams), pointer :: sourcePars(:)
    integer, intent(out) :: nSourcePars

    real, allocatable :: image(:, :), background(:, :), newImage(:, :, :)
    integer, allocatable :: levelImage(:, :)
    integer :: newNx, newNy
    integer :: i
    integer :: allocErr

    !************************************************
    ! Create new images that just contain the source
    ! of interest 
    !************************************************ 

    newNx = xhigh - xlow + 1
    newNy = yhigh - ylow + 1

    allocate(image(newNx, newNy), stat=allocErr)
    call CheckAllocation(allocErr, "image") 
 
    image = inImage(xlow : xhigh, ylow : yhigh)

    allocate(background(newNx, newNy), stat=allocErr)
    call CheckAllocation(allocErr, "background") 
 
    background = inBackground(xlow : xhigh, ylow : yhigh)

    allocate(levelImage(newNx, newNy), stat=allocErr)
    call CheckAllocation(allocErr, "levelImage") 
 
    levelImage = inLevelImage(xlow : xhigh, ylow : yhigh)

  
    !********************************
    ! Subtract the background
    !*******************************
    image = image - background

  
    if(newNx .gt. maxXsize .and. newNy .gt. maxYsize) then

        where(levelImage .ne. srcNo)
            levelImage = 0
        end where
        !**************************************************
        ! Full deblend of potentially overlapping sources
        !**************************************************
        call getSourceParameter(levelimage, image, background, newNx, newNy,&
            sourcePars, nSourcePars, nSigma, 2.0, fraction, srcNo)

        if(nSourcePars .eq. 0) then

            ! extended faint source?
            nSourcePars = 1
            allocate(sourcePars(1), stat=allocErr)
            call CheckAllocation(allocErr, "sourcePars(1) (detectPars)") 
 
            allocate(newImage(newNx, newNy, 1), stat=allocErr)
            call CheckAllocation(allocErr, "newImage (detectPars)") 
 

            !***************************************************
            ! Produce an image (newImage) of source number srcNo
            !***************************************************
            where(levelImage .eq. srcNo)
                newImage(1 : newNx, 1 : newNy, 1) = image
            elsewhere
                newImage(1 : newNx, 1 : newNy, 1) = pixelIgnoreFlag   ! flag for ingorning pixels
            end where
            write(errstr,'(''Warning: Problem with deblend of Extended source : assuming single source'')')
            call message(AppMsg, Verbose)  ! Verbose was 2
            call getParametersfromMoments(newImage, background, &
                  newNx, newNy, 1, sourcePars, nSigma)
            deallocate(newImage, stat = allocErr)
            call CheckDeallocation(allocErr, "newImage") 
 
        end if

    else

        nSourcePars = 1
        allocate(sourcePars(1), stat = allocErr)
        call CheckAllocation(allocErr, "sourcePars(1) (detectPars)") 
 
        allocate(newImage(newNx, newNy, 1), stat=allocErr)
        call CheckAllocation(allocErr, "newImage (detectPars)") 
  

        !***************************************************
        ! Produce an image (newImage) of source number srcNo
        !***************************************************
        where(levelImage .eq. srcNo)
            newImage(1 : newNx, 1 : newNy, 1) = image
        elsewhere
            newImage(1 : newNx, 1 : newNy, 1) = pixelIgnoreFlag   ! flag for ingorning pixels
        end where

        call getParametersfromMoments(newImage, background, &
                       newNx, newNy, 1, sourcePars, nSigma)
        deallocate(newImage, stat = allocErr)
        call CheckDeallocation(allocErr, "newImage")
 
    end if

    deallocate(levelImage, stat = allocErr)
    call CheckDeallocation(allocErr, "levelImage")
    deallocate(image, stat = allocErr)
    call CheckDeallocation(allocErr, "image")
 
    deallocate(background, stat = allocErr)
    call CheckDeallocation(allocErr, "background")
 
    !********************************************
    ! Set to correct coordinate system in Pixels
    !********************************************
    forall (i = 1 : nSourcePars)
        sourcePars(i)%x = sourcePars(i)%x + xlow - 1
        sourcePars(i)%y = sourcePars(i)%y + ylow - 1
    end forall

end subroutine detectParams
!***********************************************
! Subroutine getSourceParameter
! Purpose- For source number srcNo calculates the
! source parameters.
! Parameters: 
! levelImage - the image of the source - each pixel
! associated with the source has its pixel value set 
! to the source number.
! image - the input image
! background - the background image
! nx & ny, the dimensions of the images.
! sourcePars - a structure to store the parameters
! of the source.
! nsourcePars - 
! nSigma -
! Dvalue -
! fraction -
! scrNo - the source under investigation. 
!***********************************************
subroutine getSourceParameter(levelImage, image, background, &
         nx, ny, sourcePars, nSourcePars, nSigma, &
         Dvalue, fraction, srcNo)

    integer, intent(in) :: nx, ny
    integer, intent(in) :: levelImage(nx, ny)
    real, intent(inout) :: image(nx, ny)
    real, intent(in)    :: background(nx, ny)
    real, intent(in)    :: nSigma
    real, intent(in)    :: Dvalue
    real, intent(in)    :: fraction
    integer, intent(in) :: srcNo

    type(sourceParams), pointer :: sourcePars(:)
    integer, intent(out) :: nSourcePars

    real, allocatable :: inImage(:, :)
    integer :: allocErr
    real :: threshStep, threshLevel
    integer :: boxsize
    type(treeStruct), pointer :: Tree
    real :: inDvalue
    real :: totalCounts

    integer :: level

    boxsize = 1
    nSourcePars = 0
    nullify(sourcePars)
  
    threshLevel = sqrt(background(nx / 2, ny / 2)) * nSigma
    threshStep = MAXVAL(image) / 16.0
    if(Dvalue .ne. 1) then
        inDvalue = (threshStep) ** (1.0 / 15.0)
    else
        inDvalue = 1
    end if

    !*******************************************
    ! Do first level - allocate Tree structure
    !*******************************************

    nullify(Tree)
    allocate(Tree, stat=allocErr)
    call CheckAllocation(allocErr, "Tree (getSourceParameter)") 

    nullify(Tree%next)
    nullify(Tree%downLevel)
    nullify(Tree%upLevel)

    !********************************************************************
    ! go round array getting pixels and sources checking for connectivity
    ! First only get those pixels that satisfy minimum criterion
    ! Step down the 16 levels for each of the detected source areas
    ! Set up whole image size - should then propogate down
    !********************************************************************
    Tree%minX = 1
    Tree%maxX = nx
    Tree%minY = 1
    Tree%maxY = ny
  
    !***************************************************************
    ! Setup pixel array at this level - this is basically all pixels
    !***************************************************************
    allocate(Tree%levelImage(nx, ny), stat=allocErr)
    call CheckAllocation(allocErr, "Tree%levelImage (getSourceParamater)")
 

    Tree%levelImage = levelImage

    allocate(inImage(nx, ny), stat=allocErr)
    call CheckAllocation(allocErr, "inImage (getSourceParamater)")


    inImage = image
    where(levelImage .gt. 0)
        inImage = image
    elsewhere
        inImage = 0
    endwhere

    totalCounts = SUM(inImage)

    level = 1

    call TreeLoop(inImage, 1, nx, 1, ny, threshLevel, threshStep, Tree, &
           boxsize, inDvalue, level, nx, ny, totalCounts, fraction)

    deallocate(inImage, stat = allocErr)
    call CheckDeallocation(allocErr, "inImage (getSourceParamater)")

    !************************************************
    ! now write out each level to see if its working
    !************************************************
    call AnalyseTree(Tree%downlevel, nx, ny, image, background, levelImage, &
        sourcePars, nSourcePars, nSigma, fraction, srcNo)

end subroutine getSourceParameter
!*************************************************************
! Check sources - if one source lies within other source then 
! this is coincidence loss so two sources are in fact the same
!*************************************************************
subroutine CheckSources(Tree)

  type(treeStruct), pointer :: Tree
  type(treeStruct), pointer :: Tree1, Tree2
  type(treeStruct), pointer :: newTree, TreePos, TreePosNew, prevTree

  integer :: i, j, k, l
  integer :: nCopy
  integer :: allocErr

  i = 0
  Tree1 => Tree%downlevel

  checkLoop1: do 
     Tree2 => Tree%downlevel
     prevTree => Tree%downlevel
     j = 0
     checkLoop2: do 
        if(i .ne. j) then
           if(Tree2%minX .ge. Tree1%minX .and. &
                Tree2%maxX .le. Tree1%maxX .and. &
                Tree2%minY .ge. Tree1%minY .and. &
                Tree2%maxY .le. Tree1%maxY) then
                
              !*****************************************************
              ! Possible coincidence loss source
              ! For moment assume this (could check to see if really
              ! surrounded)
              !*****************************************************

              do k = Tree2%minY, Tree2%maxY
                 do l = Tree2%minX, Tree2%maxX

                    if(Tree2%levelImage(l, k) .gt. 0) then                       
                       Tree1%levelImage(l, k) = 1
                    end if
                 end do
              end do
              !***************************************
              ! Now restructure Tree to remove Tree2
              !**************************************

              k = 0
              nCopy = 0
              TreePos => Tree%downlevel

              fillTree: do
                 if(k .ne. j) then  ! only copy those we want
                    nCopy = nCopy + 1
                    !*****************************
                    ! Copy data to new structure
                    !****************************
                    if(nCopy .eq. 1) then

                       allocate(newTree, stat=allocErr)
                       call CheckAllocation(allocErr, "newTree (checkSources)")
                       
                       TreePosNew => newTree
                    else

                       allocate(TreePosNew%next, stat=allocErr)
                       call CheckAllocation(allocErr, "TreePosNew%next (checkSources)")
                       
                       TreePosNew => TreePosNew%next
                    end if
                    TreePosNew%nSources = TreePos%nSources
                    TreePosNew%level = TreePos%level
                    TreePosNew%minX = TreePos%minX
                    TreePosNew%maxX = TreePos%maxX
                    TreePosNew%minY = TreePos%minY
                    TreePosNew%maxY = TreePos%maxY


                    allocate(TreePosNew%levelImage(TreePos%minX : TreePos%maxX, &
                         TreePos%minY : TreePos%maxY), stat=allocErr)
                    call CheckAllocation(allocErr, "levelImage (checkSources)")
                   
                    TreePosNew%levelImage = TreePos%levelImage
                    nullify(TreePosNew%next)
                    nullify(TreePosNew%downlevel)
                    TreePosNew%upLevel => Tree
                    TreePosNew => TreePosNew%next
                 end if

                 k = k + 1
                 TreePos => TreePos%next
              end do fillTree
              !*******************************************************
              ! Now have corrected list in TreePosNew
              ! Clear old Tree from this point and transfer pointer to 
              ! new Tree
              !******************************************************

              write(*, *) 'Supposed never to get here'
              if(j .eq. 0) then ! At top level - replace Tree%downlevel
                 call ClearTree(Tree%downlevel)
                 nullify(Tree%downlevel)
                 if(nCopy .gt. 0) Tree%downlevel => newTree
                 
              else
                 call ClearTree(prevTree%next)
                 nullify(prevTree%next)
                 if(nCopy .gt. 0) prevTree%next => newTree
              end if
           end if
        end if
        prevTree => Tree2
        Tree2 => Tree2%next
        if(.not. associated(Tree2)) exit checkLoop2
        j = j + 1
     end do checkLoop2

     Tree1 => Tree1%next
     if(.not. associated(Tree1)) exit checkLoop1
     i = i + 1
  end do checkLoop1

end subroutine CheckSources
!*******************************************
! Go down 15 levels from start value
! Must have start level already defined
!******************************************
recursive subroutine TreeLoop(image, minxIm, maxxIm, minyIm, maxyIm, &
     threshLevel, threshStep, Tree, boxsize, Dvalue, level, nx, ny,  &
     totalCounts, fraction)

  integer, intent(in) :: minxIm, maxxIm, minyIm, maxyIm
  real, intent(in) :: image(minxIm : maxxIm, minyIm : maxyIm)
  real, intent(in) :: threshLevel, threshStep
  type(treeStruct), pointer :: Tree
  integer, intent(in) :: boxsize
  real, intent(in) :: Dvalue
  integer, intent(in) :: level
  integer, intent(in) :: nx,ny
  real, intent(in) :: totalCounts
  real, intent(in) :: fraction

  integer :: allocErr
  type(treeStruct), pointer :: nextTree
  integer, allocatable :: levelImage(:, :)
  integer :: minX, maxX, minY, maxY
  real :: IThreshLevel
  integer :: xsize,ysize
  integer :: i, j
  integer :: dataThere
  real, allocatable :: newImage(:, :)


  !*********************************************************************
  ! Check size of array to be searched - must be at least > 10 pixels in
  ! one dimension (to resolve multiple sources)
  ! Do it using pixel lists rather than arrays - keeps with sources
  ! Go down pixel list for those that are above the threshold and
  ! check connectivity to get sources
  ! Put pixels into array        
  ! Use min/max values to setup up just enough memory
  !********************************************************************

  if (any(Tree%levelImage .gt. 0)) then
     dataThere = 1
  else
     dataThere = 0
  end if


  if(dataThere .gt. 0) then  ! havn't reached the top yet

     ! Now have image of pixels for this source at this level

     allocate(levelImage(Tree%minX : Tree%maxX, Tree%minY : Tree%maxY), &
          stat = allocErr)
     call CheckAllocation(allocErr, "levelImage (TreeLoop)")
     

     ! Get data a new threshold

     if (Dvalue .eq. 1) then
        IThreshLevel = threshLevel + level * threshStep
     else
        IThreshLevel = threshLevel * Dvalue ** level
     end if

     do i = Tree%minY, Tree%maxY
        do j = Tree%minX, Tree%maxX
!	print *,j,i
	if(image(j, i) .gt. IThreshLevel) then

              levelImage(j, i) = 1
           else
              levelImage(j, i) = 0
           end if
        end do
     end do

     ! Get size and check to see if we can actually look for more than 1 source

     xsize = Tree%maxX - Tree%minX + 1
     ysize = Tree%maxY - Tree%minY + 1

     ! Note that area must have some width	


     if ((xsize .gt. maxXsize .or. ysize .gt. maxYsize) .and. (xsize .gt. 2) .and. &
	(ysize .gt. 2)) then	


        ! Here we try to split out confused sources
 
        ! Allocate next branch of Tree

        call TreeStep(image, levelImage, Tree%minX, Tree%maxX, Tree%minY, &
             Tree%maxY, Tree%downlevel, Tree, boxsize, level, totalCounts, fraction)

        ! Check to see if we're two unconnected sources inside one
        ! another - ie. photon coincidence loss source

!        call CheckSources(Tree)
     else

        ! Just propograte OK pixels into Tree 
        ! This is so know how bright feature is...
        
        allocate(Tree%downlevel, stat=allocErr)
        call CheckAllocation(allocErr, "Tree%downlevel (TreeLoop)")
        
        
        ! Setup tree structure

        nextTree => Tree%downlevel
        nullify(nextTree%next)
        nullify(nextTree%downLevel) 
        nextTree%upLevel => Tree
        nextTree%nSources = 1
        minX = Tree%maxX + 1
        minY = Tree%maxY + 1
        maxX = Tree%minX - 1
        maxY = Tree%minY - 1

	! Deallocate level image and resize
	
	deallocate(levelimage, stat = allocErr)
        call CheckDeallocation(allocErr, "levelimage (TreeLoop)")

        allocate(levelImage(minxIm : maxxIm, minyIm : maxyIm), &
                 stat=allocErr)
        call CheckAllocation(allocErr, "levelImage (TreeLoop)")
       
	

        ! Add in relavant pixels into pixel list
        
        levelImage = 0

        do i = minyIm, maxyIm
           do j = minxIm, maxxIm
              if(image(j, i) .gt. IThreshLevel) then

                 levelImage(j, i) = 1
                 minX = MIN(j, minX)
                 minY = MIN(i, minY)
                 maxX = MAX(j, maxX)
                 maxY = MAX(i, maxY)
              end if
           end do
        end do

        allocate(nextTree%levelImage(minX : maxX, minY : maxY), stat=allocErr)
        call CheckAllocation(allocErr, "nextTree%levelImage (TreeLoop)")
        
        nextTree%levelImage = levelImage(minX : maxX, minY : maxY)
        nextTree%minX = minX
        nextTree%maxX = maxX
        nextTree%minY = minY
        nextTree%maxY = maxY

     end if
        
     ! Now go down levels if necessray

     if(level .lt. 15) then
        
        nextTree => Tree%downlevel
        loop: do 
           
           if(associated(nextTree)) then
              
              !******************************************************
              ! Some pixels which can be further split
              ! Make image using levelimage so concentrate on what is
              ! important
              !*****************************************************
              allocate(newImage(nextTree%minX : nextTree%maxX, &
                   nextTree%minY : nextTree%maxY), stat=allocErr)
              call CheckAllocation(allocErr, "newImage (TreeLoop)")
             

              ! Transger image data over - only those pixels that are needed

              newImage = 0
              forall (i = nextTree%minY : nextTree%maxY)
                 forall(j = nextTree%minX: nextTree%maxX) newImage(j, i) = image(j, i)
              end forall

              call TreeLoop(newImage, nextTree%minX, nextTree%maxX, &
                   nextTree%minY, nextTree%maxY, threshLevel, threshStep, &
                   nextTree, boxsize, Dvalue, level + 1, nx, ny, totalCounts, fraction)
              deallocate(newImage, stat = allocErr)
              call CheckDeallocation(allocErr, "newImage (TreeLoop)")

           else
     
              exit loop
              
           endif
           
           nextTree => nextTree%next
           
        end do loop
     
     end if
     
     deallocate(levelImage, stat = allocErr)
     call CheckDeallocation(allocErr, "levelImage (TreeLoop)")

  end if


end subroutine TreeLoop
!*****************************************************************
! Split out the individual sources at this level and store in Tree
!*****************************************************************
subroutine TreeStep(realImage, image, minx, maxx, miny, maxy, Tree, upTree, &
     boxsize, level, totalCounts, fraction)

  integer, intent(in) :: minX, maxX, minY, maxY
  real, intent(in) :: realImage(minX : maxX, minY : maxY)
  integer, intent(inout) :: image(minX : maxX, minY : maxY)
  type(TreeStruct), pointer :: Tree
  type(TreeStruct), pointer :: upTree
  integer, intent(in) :: boxsize
  integer, intent(in) :: level
  real, intent(in) :: totalCounts
  real, intent(in) :: fraction

  integer :: nSource
  integer :: maxpos(2)
  integer :: xpos, ypos
  integer :: allocErr
  type(TreeStruct), pointer :: nextTree
  integer, allocatable :: levelImage(:, :)
  integer :: xmin, xmax, ymin, ymax
  real :: sourceCounts

  ! Allocate levelImage

  allocate(levelImage(minX : maxX, minY : maxY), stat=allocErr)
  call CheckAllocation(allocErr, "levelImage (TreeStep[A)")
  

  ! Make each of the branches at this level

  nSource = 0

  splitSources: do
     
     ! Check connectivity start from top as a place to start
     
     maxpos = MAXLOC(image)
     xpos = maxpos(1) + minX - 1
     ypos = maxpos(2) + minY - 1
     if(image(xpos, ypos) .gt. 0) then

        ! First get level image 
        
        xmin = xpos
        xmax = xpos
        ymin = ypos
        ymax = ypos
        sourceCounts = image(xpos, ypos)
        levelImage = 0


        levelImage(xpos, ypos) = 1
	call nearestNeighbourNewParam(realImage, image, levelImage, &
             minX, maxX, minY, maxY, xmin, xmax, ymin, ymax, xpos, ypos, boxsize, &
             sourceCounts)

        ! now check to see if this source should be included 
        ! ie. sourceCOunts > fraction*totalCoumts

	if(sourceCounts .ge. fraction * totalCounts) then

           nSource = nSource + 1

           if(nSource .eq. 1) then
              allocate(Tree, stat=allocErr)
              call CheckAllocation(allocErr, "Tree (TreeStep)")
              
              nextTree => Tree
              nullify(nextTree%next)
              nullify(nextTree%downlevel)
              nextTree%upLevel => upTree
           else
              ! Allocate next Tree level
              allocate(nextTree%next,stat=allocErr)
              call CheckAllocation(allocErr, "nextTree%next (TreeStep)")
             
              nextTree => nextTree%next
              nullify(nextTree%next)
              nullify(nextTree%downlevel)
              nextTree%upLevel => upTree
           end if

        ! First get where connected pixels are in levelImage

       
        ! Put into tree

           nextTree%nSources = 1
           nextTree%level = level
           nextTree%minX = xmin
           nextTree%maxX = xmax
           nextTree%minY = ymin
           nextTree%maxY = ymax
           
           ! Transfer this source into Tree
           
           allocate(nextTree%levelImage(xmin : xmax, ymin : ymax), stat = allocErr)
           call CheckAllocation(allocErr, "nextTree%levelImage (TreeStep)")
          
           nextTree%levelImage(xmin : xmax, ymin: ymax) = &
                levelImage(xmin : xmax, ymin : ymax)

        end if

        ! put values accessed to zero in image
        
        where(levelImage .gt. 0) image = 0
       
     else
        
        exit splitSources
        
     end if

  end do splitSources

  if(nSource .gt. 0) then
     if(associated(upTree)) then
        upTree%nSources = nSource
     else
        Tree%nSources = nSource
     end if
  end if

  deallocate(levelImage, stat = allocErr)
  call CheckDeallocation(allocErr, "levelImage (TreeStep)")


end subroutine TreeStep
!*******************************************
! nearestNeighbourNewParam
!*******************************************
subroutine nearestNeighbourNewParam(realImage, newImage, levelImage, &
     minX, maxX, minY, maxY, xmin, xmax, ymin, ymax, &
     xpos, ypos, boxsize, sourceCounts)

  integer, intent(in) :: minX, maxX, minY, maxY
  real, intent(in) :: realImage(minX : maxX, minY : maxY)
  integer, intent(in) :: newImage(minX : maxX, minY : maxY)
  integer, intent(inout) :: levelImage(minX : maxX, minY : maxY)
  integer, intent(inout) :: xmin, xmax, ymin, ymax
  integer, intent(in) :: xpos, ypos
  integer, intent(in) :: boxsize
  real, intent(inout) :: sourceCounts

  type(srcPixel), pointer :: checkPixel
  type(srcPixel), pointer :: nextPixel
  type(srcPixel), pointer :: startPixel
  integer :: allocErr
  integer :: k,l
  integer :: minXpos, minYpos, maxXpos, maxYpos

  ! Do middle line

  allocate(startPixel, stat=allocErr)
  call CheckAllocation(allocErr, "pixel%next (nearestNeighbourNewParam)")
  

  startPixel%x = xpos
  startPixel%y = ypos
  nullify(startPixel%next)
  nullify(startPixel%previous)
  checkPixel => startPixel
  nextPixel => startPixel

  loop: do

     call getMinMaxPositionParam(checkPixel%x, checkPixel%y, &
          minX, maxX, minY, maxY, minXpos, maxXpos, minYpos, maxYpos, boxsize)
     
     do k = minYpos, maxYpos
        do l = minXpos, maxXpos
           if (k .ne. checkPixel%x .and. l .ne. checkPixel%y) then

              if(levelImage(l , k) .eq. 0) then
                 if(newimage(l , k) .gt. 0) then
                    levelImage(l , k) = 1

                    xmin = MIN(l, xmin)
                    xmax = MAX(l, xmax)
                    ymin = MIN(k, ymin)
                    ymax = MAX(k, ymax)

                    sourceCounts = sourceCounts + realImage(l, k)

                    allocate(nextPixel%next, stat=allocErr)
                    call CheckAllocation(allocErr, "pixel%next (nearestNeighbourNewParam)")
                    
                    
                    nextPixel%next%previous => nextPixel
                    nextPixel => nextPixel%next
                    nextPixel%x = l
                    nextPixel%y = k
                    nullify(nextPixel%next)
                 end if
              end if
           end if
        end do
     end do

     ! Move check position - if come to end exit

     if(associated(checkPixel%next)) then
        checkPixel => checkPixel%next
     else
        exit loop
     end if

  end do loop

  ! Free up link list

  checkPixel => startPixel
  downPixelList: do

     if(associated(checkPixel%next)) then
        checkPixel => checkPixel%next
     else
        exit downPixelList
     end if

  end do downPixelList

  upPixelList: do

     if(associated(checkPixel%previous)) then
        checkPixel => checkPixel%previous
        deallocate(checkPixel%next, stat = allocErr)
        call CheckDeallocation(allocErr, "checkPixel%next (nearestNeighbourNewParam)")

        nullify(checkPixel%next)
     else
        exit upPixelList
     end if
  end do upPixelList
  deallocate(startPixel, stat = allocErr)
  call CheckDeallocation(allocErr, "startPixel (nearestNeighbourNewParam)")

  nullify(startPixel)

end subroutine nearestNeighbourNewParam
!**************************************************
! getMinMaxPositionParam
!*************************************************
subroutine getMinMaxPositionParam(xpos, ypos, minX, maxX, minY, maxY, &
     minXpos, maxXpos, minYpos, maxYpos, boxsize)

  integer, intent(in) :: xpos, ypos
  integer, intent(in) :: minX, minY, maxX, maxY
  integer, intent(out) :: minXpos, maxXpos
  integer, intent(out) :: minYpos, maxYpos
  integer, intent(in) :: boxsize

  if (xpos - boxsize .ge. minX .and. xpos + boxsize .le. maxX) then
     minXpos = xpos - boxsize
     maxXpos = xpos + boxsize
  else if (xpos .eq. minX) then
     minXpos = minX
     maxXpos = MIN(minX + boxsize, maxX)
  else
     minXpos = MAX(maxX - boxsize, minX)
     maxXpos = maxX
  end if

  if(ypos - boxsize .ge. minY .and. ypos + boxsize .le. maxY) then
     minYpos = ypos - boxsize
     maxYpos = ypos + boxsize
  else if(ypos .eq. minY) then
     minYpos = minY
     maxYpos = MIN(minY + boxsize, maxY)
  else
     minYpos = MAX(maxY - boxsize, minY)
     maxYpos = maxY
  end if

end subroutine getMinMaxPositionParam
!**************************************************
! nextToPosition
!**************************************************
recursive subroutine nextToPosition(realImage, newImage, levelImage, &
     minX, maxX, minY, maxY, xmin, xmax, ymin, ymax, xpos, ypos, boxsize, sourceCounts)

  integer, intent(in) :: minX, maxX, minY, maxY
  real, intent(in) :: realImage(minX : maxX, minY : maxY)
  integer, intent(in) :: newImage(minX : maxX, minY : maxY)
  integer, intent(inout) :: levelImage(minX : maxX, minY : maxY)
  integer, intent(inout) :: xmin, xmax, ymin, ymax
  integer, intent(in) :: xpos, ypos
  integer, intent(in) :: boxsize
  real, intent(inout) :: sourceCounts

  integer :: minXpos, maxXpos, minYpos, maxYpos
  integer :: i, j
  
  minXpos = MAX(minX, xpos - boxsize)
  minYpos = MAX(minY, ypos - boxsize)
  maxXpos = MIN(maxX, xpos + boxsize)
  maxYpos = MIN(maxY, ypos + boxsize)

  do i = minYpos, maxYpos
     do j = minXpos, maxXpos
        if(i .ne. xpos .or. j .ne. ypos) then

           if(levelImage(j, i) .eq. 0) then
              if(newImage(j, i) .gt. 0) then
                 levelImage(j, i) = 1
                 xmin = MIN(j, xmin)
                 xmax = MAX(j, xmax)
                 ymin = MIN(i, ymin)
                 ymax = MAX(i, ymax)
                 sourceCounts = sourceCounts + realImage(j, i)
                 call nextToPosition(realImage, newImage, levelImage, minX, maxX, &
                      minY, maxY, xmin, xmax, ymin, ymax, j, i, boxsize, &
                      sourceCounts)
              end if
           end if
        end if
     end do
  end do

end subroutine nextToPosition
!******************************************************************
! Routine to get number of sources so can store in array of sources
! This created a simple tree with just the nodes stored.
!******************************************************************
recursive subroutine GetToEndOfTree2(Tree, outTree, flag, nx, ny, testim)

  type(treeStruct), pointer :: Tree
  type(treeStruct), pointer :: outTree
  integer, intent(inout) :: flag
  integer, intent(in) :: nx,ny
  integer, intent(inout) :: testim(nx,ny)

  type(treeStruct), pointer :: nextTree
  type(treeStruct), pointer :: upTree

  integer :: allocErr
 
  integer :: upperBound(2)

  nextTree => Tree
  loop: do
     if(associated(nextTree)) then
        if(nextTree%level .ge. 0) then  ! if -1/-2 already been here
           if(associated(nextTree%downlevel)) then
              call GetToEndOfTree2(nextTree%downlevel, outTree, flag, nx, ny, testim)
           else
!              print *,"End of branch",nextTree%minX,nextTree%maxX,nextTree%minY,nextTree%maxY
              flag = 1 ! found one!
              upTree => nextTree%upLevel
              innerLoop: do
                 if(associated(upTree)) then
                    if(upTree%nSources .gt. 1) then

   
                       upperBound = ubound(nextTree%levelImage)

                       if(upperBound(2) .eq. 0 .and. upperBound(2) .eq. 0) then

                          nextTree => nextTree%upLevel
                          upperBound = ubound(nextTree%levelImage)
                          if(upperBound(2) .eq. 0 .and. upperBound(2) .eq. 0) exit loop
                          
                       end if
                       outTree%level = nextTree%level
                       outTree%minX = nextTree%minX
                       outTree%maxX = nextTree%maxX
                       outTree%minY = nextTree%minY
                       outTree%maxY = nextTree%maxY
                       !                       print *,"level :",upTree%level
                       !                       print *,"End of branch",outTree%minX,outTree%maxX,outTree%minY,outTree%maxY
                       allocate(outTree%levelImage(outTree%minX : outTree%maxX, &
                            outTree%minY : outTree%maxY),stat = allocErr)
                       call CheckAllocation(allocErr, "outTree%levelImage (getToEndOfTree2)")
                      
                       outTree%levelImage = nextTree%levelImage
                       
                       ! Allocate next one - note level = -1 so we can spot 
                       ! end of this list
                       
                       allocate(outTree%next, stat=allocErr)
                       call CheckAllocation(allocErr, "outTree%next (getToEndOfTree2)")
                     
                       outTree%next%level=-1
                       nullify(outTree%next%downLevel)
                       nullify(outTree%next%upLevel)
                       nullify(outTree%next%levelImage)
                       nullify(outTree%next%next)
                       outTree => outTree%next
                       nextTree%level = -1
                       exit loop
                    else
                       nextTree => upTree
                       upTree=>upTree%upLevel
                    end if
                 else
                    nextTree%level = -1
                    exit loop
                 endif
              end do innerLoop
           endif
        endif
     else        
        exit loop
     endif
     nextTree => nextTree%next
  end do loop
  
end subroutine GetToEndOfTree2
!*******************************************************
! GetToEndOfTree
!*******************************************************

recursive subroutine GetToEndOfTree(Tree, outTree)

  type(treeStruct), pointer :: Tree
  type(treeStruct), pointer :: outTree

  type(treeStruct), pointer :: nextTree
  type(treeStruct), pointer :: upTree


  integer :: allocErr

  nextTree => Tree
  bigLoop: do 
     if(.not. associated(nextTree)) then
        exit bigLoop
     end if
     if(associated(nextTree%downlevel)) then
        call GetToEndOfTree(nextTree%downlevel, outTree)     
     else
        ! Analyse this one ie. get to branch position
        upTree => nextTree%upLevel  ! do up in tree
        loop: do
           if(associated(upTree)) then
!              print *,"*",upTree%upLevel%nSources,upTree%minX,&
!                   upTree%maxX,upTree%minY,upTree%maxY
              if(upTree%nSources .gt. 1) then
                 ! nextTree is start of branch
                 ! copy into this tree to ease bookeeping
                 outTree%level = nextTree%level
                 outTree%minX = nextTree%minX
                 outTree%maxX = nextTree%maxX
                 outTree%minY = nextTree%minY
                 outTree%maxY = nextTree%maxY
!                 print *,"level :",upTree%level
!                 print *,"End of branch",outTree%minX,outTree%maxX,outTree%minY,outTree%maxY
                 allocate(outTree%levelImage(outTree%minX : outTree%maxX, &
                      outTree%minY : outTree%maxY),stat = allocErr)
                 call CheckAllocation(allocErr, "outTree%levelImage (getToEndOfTree)")
                
                 outTree%levelImage = nextTree%levelImage

                 ! Allocate next one - note level = -1 so we can spot end of
                 ! this list

                 allocate(outTree%next, stat=allocErr)
                 call CheckAllocation(allocErr, "outTree%next (getToEndOfTree)")
                 
                 outTree%next%level = -1
                 nullify(outTree%next%downLevel)
                 nullify(outTree%next%upLevel)
                 nullify(outTree%next%levelImage)
                 nullify(outTree%next%next)
                 outTree => outTree%next
                 exit loop
              end if
           else ! nextTree is start of branch since at top level already!
              outTree%level = upTree%level
              outTree%minX = upTree%minX
              outTree%maxX = upTree%maxX
              outTree%minY = upTree%minY
              outTree%maxY = upTree%maxY
              allocate(outTree%levelImage(outTree%minX : outTree%maxX, &
                   outTree%minY : outTree%maxY),stat=allocErr)
              call CheckAllocation(allocErr, "outTree%levelImage (getToEndOfTree)")
              
              outTree%levelImage = upTree%levelImage
              allocate(outTree%next, stat=allocErr)
              call CheckAllocation(allocErr, "outTree%next (getToEndOfTree)")
             
              outTree%next%level=-1
              nullify(outTree%next%downLevel)
              nullify(outTree%next%upLevel)
              nullify(outTree%next%levelImage)
              nullify(outTree%next%next)
              outTree => outTree%next
              exit loop
           end if
           upTree => upTree%upLevel  ! do up in tree
        end do loop
     endif
     nextTree => nextTree%next
  end do bigLoop

end subroutine GetToEndOfTree
!******************************************
! AnalyseTree
!******************************************

subroutine AnalyseTree(Tree, nx, ny, image, background, levelImage, srcParameters, &
     nUseSource, nSigma, fraction, srcNo)

  type(treeStruct), pointer :: Tree
  integer, intent(in) :: nx, ny
  real, intent(in) :: image(nx, ny)
  real, intent(in) :: background(nx, ny)
  integer,intent(in) :: levelImage(nx, ny)
  type(sourceParams), pointer :: srcParameters(:)
  integer, intent(out) :: nUseSource
  real, intent(in) :: nSigma
  real, intent(in) :: fraction
  integer, intent(in) :: srcNo

  type(treeStruct), pointer :: nextTree
  type(treeStruct), pointer :: topTree
  type(treeStruct), pointer :: outTree
  integer :: numberSources
  integer :: allocErr
  integer :: level

  real :: totalCounts
  real, pointer :: fitParams(:,:)
  integer, pointer :: xRange(:,:), yRange(:,:)
  integer :: flag
  integer, allocatable :: testim(:,:)
  real, allocatable :: newImage(:,:)
  
  ! Variables about each source to split out pixels so really calculate
  ! fluxes etc.


  ! Go to end of branches and work backwards

  ! Need at each point to check two levels down 
  ! Deletes branches as we go down

  ! Go to bottom level

  nullify(topTree)
  if (.not. associated(Tree)) then

     ! No tree exists

     nUseSource = 0
     return
  end if
  if (.not. associated(Tree%downlevel)) then

     ! No tree exists

     nUseSource = 0
     return
  end if
  nextTree => Tree

  ! Get total counts

  call checkSourceCounts(totalCounts, 1, nx, 1, ny, nx, ny, image, levelImage)

  ! Find out how many nodes there are ie. confused sources

  allocate(topTree, stat=allocErr)
  call CheckAllocation(allocErr, "topTree (AnalyseTree)")
 
  nullify(topTree%downlevel)
  nullify(topTree%uplevel)
  nullify(topTree%next)
  nullify(topTree%levelImage)
  numberSources = 0
  outTree => topTree
  level = 1
  call zeroTree(nextTree)

  ! Fix tree and remove spurious sources at this point

  flag = 0
  call fixTree(nextTree, nextTree, level, nx, ny, image)
!  print *,"ENDOFFIX"
  nullify(nextTree%upLevel)
  allocate(testim(nx, ny), stat=allocErr)
  call CheckAllocation(allocErr, "test (AnalysisTree)") 
 
  loop: do
     flag = 0
     call GetToEndOfTree2(nextTree, outTree, flag, nx, ny, testim)
     if(flag .eq. 0) exit loop
  end do loop
  deallocate(testim, stat = allocErr)
  call CheckDeallocation(allocErr, "testim (AnalysisTree)") 

  call clearTree(Tree)

  ! get fits

  call getFits(image, nx, ny, nUseSource, topTree, totalCounts, fraction, &
               fitParams, xRange, yRange)

  ! Now many sources are there in topTree
  call clearTree(topTree)

  nullify(srcParameters)

  if(nUseSource .gt. 0) then
     
     ! Only worry about confustion if confused...

!     if(srcNo == 12)then
!       write(errstr,'(''At source number 12 : nsources = '',i5)')nUseSource
!       call message(AppMsg, 2)
!     endif
     if(nUseSource .gt. 1)then
        call confusedSourceParams(image, background, nx, ny, &
             nUseSource, fitParams, srcParameters, nSigma)
     else
        allocate(srcParameters(1), stat=allocErr)
        call CheckAllocation(allocErr, "srcParameters (Analyse tree)")
        
        allocate(newImage(nx, ny), stat=allocErr)
        call CheckAllocation(allocErr, "newImage (confusedSourceParams)")
       
        where(levelImage .eq. srcNo)
           newImage = image
        elsewhere
           newImage = pixelIgnoreFlag
        end where
        call getParametersfromMoments(newImage, background, &
                                      nx, ny, 1, srcParameters, nSigma)
        deallocate(newImage, stat = allocErr)
        call CheckDeallocation(allocErr, "newImage (confusedSourceParams)")
   
    end if
    deallocate(fitParams)
       call CheckDeallocation(allocErr, "fitParams (confusedSourceParams)")
  
       deallocate(xRange, stat = allocErr)
       call CheckAllocation(allocErr, "xRange (confusedSourceParams)")
    

      deallocate(yRange, stat = allocErr)
      call CheckAllocation(allocErr, "yRange (confusedSourceParams)")
  
  end if
end subroutine AnalyseTree
!***************************************************************************
! Bug somewhere - try to fix tree here so understand number sources,level nos
! Almost certainly upTree%nSources = upTree%nsources + nsources isn't in 
! tree code - fix later
! Also remove entries which dont have enough counts at this point
! stops noise interfering with analysing tree
!***************************************************************************
recursive subroutine fixTree(Tree, upTree, level, nx, ny, image)
  
    type(treeStruct), pointer :: Tree, upTree
    integer, intent(in) :: level
    integer, intent(in) :: nx,ny
    real, intent(in) :: image(nx, ny)
    integer :: nSources
    integer :: temp
    type(treeStruct), pointer :: nextTree, prevTree

    nSources = 0
    prevTree => Tree
    nextTree => Tree

    loop: do
        if( .not. associated(nextTree)) then
        upTree%nSources = upTree%nSources + nSources
!        print *,"At level : ",level," there are ",upTree%nSources," sources"
            exit loop
        else

        ! Bug somewhere - fix it later

            if(nextTree%minX .gt. nextTree%maxX) then
                temp = nextTree%minX
                nextTree%minX = nextTree%maxX
                nextTree%maxX = temp
            end if

            if(nextTree%minY .gt. nextTree%maxY) then
                temp = nextTree%minY
                nextTree%minY = nextTree%maxY
                nextTree%maxY = temp
            end if

            nSources = nSources + 1
            nextTree%level = level

            if(associated(nextTree%downlevel)) then
                call fixTree(nextTree%downlevel, nextTree, level + 1, nx, ny, image)
            else
                nextTree%nSources = 1
            end if
        endif
        prevTree => nextTree
        nextTree => nextTree%next
    end do loop


end subroutine fixTree
!*********************************************************
! set nSources at all nodes to zero in preparation for fix
!*********************************************************
recursive subroutine zeroTree(Tree)

  type(treeStruct), pointer :: Tree
  type(treeStruct), pointer :: nextTree

  nextTree => Tree
  loop: do
     if (associated(nextTree)) then
        nextTree%nSources = 0
        if (associated(nextTree%downlevel)) then
           call zeroTree(nextTree%downlevel)
        endif
     else
        exit loop
     endif
     nextTree => nextTree%next
  end do loop

end subroutine zeroTree
!*****************************************
! getFits
! Get gaussian fits to data
!*****************************************
subroutine getFits(image, nx, ny, nUseSource, topTree, totalCounts, fraction, &
     fitParams, xRange, yRange)

  integer, intent(in) :: nx, ny
  real, intent(in) :: image(nx, ny)
  integer, intent(out) :: nUseSource
  type(treeStruct), pointer :: topTree
  real, intent(in) :: totalCounts
  real, pointer :: fitParams(:,:)
  real, intent(in) :: fraction
  integer, pointer :: xRange(:,:), yRange(:,:)

  integer :: nPixels
  real :: sourceCounts
  type(treeStruct), pointer :: outTree

  real :: gfitParams(6)
  real, allocatable :: tempParams(:,:)
  integer, allocatable :: tempXrange(:,:),tempYrange(:,:)
  integer :: allocErr
  integer :: pixXsize, pixYsize
 
  nUseSource = 0
  outTree => topTree
  checkLoop: do
     if (associated(outTree)) then
        if (outTree%level .ne. -1) then

           ! Check to see if this is a significant source > 5 sigma
           ! Get source counts and ring of `background' pixels

           pixXsize = outTree%maxX - outTree%minX + 1
	   pixYsize = outTree%maxY - outTree%minY + 1
	   nPixels = pixXsize*pixYsize
	   
           ! Since we're trying to look at complex extended structures
           ! ignore single pixels detections - really likely to be associated
           ! with extended source

           if(nPixels .gt. 1 .and. pixXsize .gt. 2 .and. pixYsize .gt. 2) then
           
              ! Make sure source has at least fraction of total counts
              
              call checkSourceCounts(sourceCounts, &
                   outTree%minX, outTree%maxX, outTree%minY, outTree%maxY, &
                   nx, ny, image, outTree%levelImage)
              if(sourceCounts .ge. fraction * totalCounts) then
                 nUseSource = nUseSource + 1

                 !***************************
                 ! fit gaussian to this data 
                 !***************************
                 call fitGaussian(image, nx, ny, outTree%levelImage, &
                      outTree%minX, outTree%maxX, outTree%minY, outTree%maxY, &
                      gfitParams)

                 if(nUseSource .gt. 1) then

                    allocate(tempParams(6,nUseSource-1),stat=allocErr)
                    call CheckAllocation(allocErr, "tempParams (AnalyseTree)")
              
                    tempParams = fitParams
                    deallocate(fitParams, stat = allocErr)
                    call CheckDeallocation(allocErr, "tempParams (AnalysisTree)")
  
                    allocate(fitParams(6, nUseSource), stat=allocErr)
                    call CheckAllocation(allocErr, "fitParams (AnalyseTree)")
                    
                    fitParams(1 : 6, 1 : nUseSource - 1) = tempParams
                    fitParams(1 : 6, nUseSource) = gfitParams
                    deallocate(tempParams, stat = allocErr)
                    call CheckDeallocation(allocErr, "tempParams (AnalysisTree)")
  

                    allocate(tempXrange(2, nUseSource - 1),stat=allocErr)
                    call CheckAllocation(allocErr, "tempXrange (AnalyseTree)")
                   
                    tempXrange = xRange
                    deallocate(xRange, stat = allocErr)
                    call CheckDeallocation(allocErr, "tempXrange (AnalysisTree)")
  
                    allocate(xRange(2, nUseSource),stat=allocErr)
                    call CheckAllocation(allocErr, "xRange (AnalyseTree)")
                    
                    xRange(1 : 2, 1 : nUseSource - 1) = tempXrange
                    xRange(1, nUseSource) = outTree%minX
                    xRange(2, nUseSource) = outTree%maxX
                    deallocate(tempXrange, stat = allocErr)
                    call CheckDeallocation(allocErr, "tempXrange (AnalysisTree)")
  
                    allocate(tempYrange(2, nUseSource - 1), stat = allocErr)
                    call CheckAllocation(allocErr, "tempYrange (AnalyseTree)")
                  
                    tempYrange = yRange
                    deallocate(yRange, stat = allocErr)
                    call CheckDeallocation(allocErr, "yRange (AnalysisTree)")
  
                    allocate(yRange(2, nUseSource), stat=allocErr)
                    call CheckAllocation(allocErr, "yRange (AnalyseTree)")
                    
                    yRange(1 : 2, 1 : nUseSource - 1) = tempYrange
                    yRange(1, nUseSource) = outTree%minY
                    yRange(2, nUseSource) = outTree%maxY
                    deallocate(tempyrange, stat = allocErr)
                    call CheckDeallocation(allocErr, "tempyrange (AnalysisTree)")
  
                 else
                    allocate(fitParams(6, nUseSource), stat=allocErr)
                    call CheckAllocation(allocErr, "fitParams (AnalyseTree)")
                    
                    fitParams(1 : 6, 1) = gfitParams
                    allocate(xRange(2, nUseSource), stat=allocErr)
                    call CheckAllocation(allocErr, "xRange (AnalyseTree)")
                   
                    xRange(1, 1) = outTree%minX
                    xRange(2, 1) = outTree%maxX
                    allocate(yRange(2, nUseSource), stat=allocErr)
                    call CheckAllocation(allocErr, "yRange (AnalyseTree)")
                    
                    yRange(1, 1) = outTree%minY
                    yRange(2, 1) = outTree%maxY
                 endif

              else
                 outTree%level = -2
              endif
           else
              outTree%level = -2
           endif
        else
           exit checkLoop
        endif
     else
        exit checkLoop
     endif
     outTree => outTree%next
  end do checkLoop

end subroutine getFits
!**********************************************
! CheckSourceCounts
!**********************************************
subroutine CheckSourceCounts(sourceCounts, minX, maxX, minY, maxY, nx, ny, image, &
     levelImage)

  real, intent(out) :: sourceCounts
  integer, intent(in) :: minX, maxX, minY, maxY
  integer, intent(in) :: nx, ny
  real, intent(in) :: image(nx, ny)
  integer, intent(in) :: levelImage(minX : maxX, minY : maxY)

  integer i, j

  sourceCounts = 0.0
  do i = minY, maxY
     do j = minX, maxX
          if(levelImage(j, i) .ne. 0) sourceCounts = sourceCounts + image(j, i)
     end do   
  end do

end subroutine CheckSourceCounts
!**********************************************************
! GetSourceBackCountsAnn
!**********************************************************
subroutine  GetSourceBackCountsAnn(sourceCounts, backCounts, &
     minX, maxX, minY, maxY, nx, ny, image, background, levelImage)

  real, intent(out) :: sourceCounts, backCounts
  integer, intent(in) :: minX, maxX, minY, maxY
  integer, intent(in) :: nx, ny
  real, intent(in) :: image(nx, ny)
  real, intent(in) :: background(nx, ny)
  integer, intent(in) :: levelImage(minX : maxX, minY : maxY)

  integer :: i, j, k, l
  integer :: minXvalue,maxXvalue
  integer :: minYvalue,maxYvalue
  logical :: useData
  real :: numBackCounts, numSourceCounts

  sourceCounts = 0
  backCounts = 0
  numSourceCounts =0
  numBackCounts =0

  do i = minY, maxY
     do j = minX, maxX
!write(*, *) "line 1654 - i=", i, "j=", j

        if(levelImage(j, i) .ne. 0) then
           sourceCounts = sourceCounts + image(j, i)
           numSourceCounts = numSourceCounts + 1
        else           
           minXvalue = MAX(j - 1, minX)
           maxXvalue = MIN(j + 1, maxX)
           minYvalue = MAX(i - 1, minY)
           maxYvalue = MIN(i + 1, maxY)
           useData = .false.
           firstLoop: do k = minYvalue, maxYvalue
              do l = minXvalue,  maxXvalue
!write(*, *) "line 1667 - k=", k, "l=", l

                 if(levelImage(l, k) .ne. 0.0) then
                    useData = .true.
                    exit firstLoop
                 end if
              end do
           end do firstLoop
           if(useData) then
              backCounts = backCounts + image(j, i)
              numBackCounts = numBackCounts + 1
           end if
        end if
     end do
  end do
  
  ! Calculate significance
              
  if(numBackCounts .gt. 0)then
     backCounts = backCounts * (numSourceCounts / numBackCounts)
  else
     do i = minY, maxY
        do j = minX, maxX
           minXvalue = MAX(j - 2, minX)
           maxXvalue = MIN(j + 2, maxX)
           minYvalue = MAX(i - 2, minY)
           maxYvalue = MIN(i + 2, maxY)
           useData = .false.
           firstLoop2: do k = minYvalue, maxYvalue
              do l = minXvalue, maxXvalue
                 if(levelImage(l, k) .ne. 0) then

                    useData = .true.
                    exit firstLoop2
                 end if
              end do
           end do firstLoop2
           if(useData) then
              backCounts = backCounts + image(j, i)
              numBackCounts = numBackCounts + 1
           end if
        end do
     end do
     backCounts = backCounts * (numSourceCounts / numBackCounts)

  end if
  if(backCounts .le. 0) then ! must use background and image - area around
                         ! source is at background so use background under
                         ! source

     backCounts = 0.0
     do i = minY, maxY
        do j = minX, maxX
           if(levelImage(j, i) .ne. 0) then
              backCounts = backCounts + background(j, i)
              numBackCounts = numBackCounts + 1
           end if
        end do
     end do

  end if

!  print *,sourceCounts,numSourceCounts,backCounts,numBackCounts
  backCounts = backCounts * (numSourceCounts / numBackCounts)


end subroutine GetSourceBackCountsAnn
!*****************************************************
! GetSourceBackCounts
!****************************************************
subroutine GetSourceBackCounts(sourceCounts, backCounts, &
     minX, maxX, minY, maxY, nx, ny, image, background, levelImage)

  real, intent(out) :: sourceCounts, backCounts
  integer, intent(in) :: minX, maxX, minY, maxY
  integer, intent(in) :: nx, ny
  real, intent(in) :: image(nx, ny)
  real, intent(in) :: background(nx, ny)
  integer, intent(in) :: levelImage(minX : maxX, minY : maxY)
  integer :: i, j
  real :: numBackCounts, numSourceCounts

  sourceCounts = 0
  backCounts = 0
  numSourceCounts =0
  numBackCounts =0

  do i = minY, maxY
     do j = minX, maxX
!write(*, *) "line 1759 - i=", i, "j=", j

        if(levelImage(j, i) .ne. 0) then
           sourceCounts = sourceCounts + image(j, i)
           numSourceCounts = numSourceCounts + 1
        else           
           backCounts = backCounts + background(j, i)
           numBackCounts = numBackCounts + 1
        end if
     end do
  end do

end subroutine GetSourceBackCounts
!*******************************************
! confusedSourceParams
!*******************************************
subroutine confusedSourceParams(image, background, nx, ny, &
     nUseSource, fitParams, srcParams, nSigma)

  integer, intent(in) :: nx, ny
  real, intent(in) :: image(nx, ny)
  real, intent(in) :: background(nx, ny)
!  integer, intent(in) :: levelImage(nx, ny)
  integer,intent(in) :: nUseSource
  real, intent(inout) :: fitParams(6, nUseSource)
!  integer, intent(in) :: xRange(2, nUseSource), yRange(2, nUseSource)
  type(sourceParams), pointer :: srcParams(:)
!  integer, intent(in) :: srcNo
  real, intent(in) :: nSigma

  integer :: allocErr

  real, allocatable :: newImage(:, :, :)
  real, allocatable :: flux(:), radius(:)
  
  allocate(srcParams(nUseSource), stat=allocErr)
  call CheckAllocation(allocErr, "srcParams (confusedSourceParams)")
  

  allocate(newImage(nx,ny,nUseSource), stat=allocErr)
  call CheckAllocation(allocErr, "newImage (confusedSourceParams)")
  
  newImage = pixelIgnoreFlag
  allocate(flux(nUseSource), stat=allocErr)
  call CheckAllocation(allocErr, "flux (confusedSourceParams)")
  
  allocate(radius(nUseSource), stat=allocErr)
  call CheckAllocation(allocErr, "radius (confusedSourceParams)")
 

  ! For each pixels assign to source dependant on 2d gaussian fit

  call getSourceImage(nx, ny, image, newImage, nUseSource,&
     flux, radius, fitParams)

  deallocate(flux, stat = allocErr)
  call CheckDeallocation(allocErr, "flux (ConfusedSourceParams)")
  
  deallocate(radius, stat = allocErr)
    call CheckDeallocation(allocErr, "radius (ConfusedSourceParams)")
 
  
  ! NewLevelImage now has all the information about the seperate
  ! sources

  call getParametersfromMoments(newImage, background, nx, ny, nUseSource, &
       srcParams, nSigma)

  ! deallocate memory

  deallocate(newImage, stat = allocErr)
  call CheckDeallocation(allocErr, "newImage (ConfusedSourceParams)")


end subroutine confusedSourceParams
!******************************************************************
! getSourceImage
!******************************************************************
subroutine getSourceImage(nx, ny, image, newImage, nUseSource, &
     flux, radius, fitParams)

  integer, intent(in) :: nx, ny
  real, intent(in) :: image(nx, ny)
!  integer, intent(in) :: levelImage(nx, ny)
  integer, intent(in) :: nUseSource
  real, intent(out) :: newImage(nx, ny, nUseSource)
  real, intent(inout) :: flux(nUseSource), radius(nUseSource)
  real, intent(in) :: fitParams(6, nUseSource)

  integer :: i, j, k
  
  real :: imageVal, fluxTotal
  real :: angle, xdist, ydist
  
  real, allocatable :: norm(:)
  integer :: allocErr

  allocate(norm(nUseSource), stat=allocErr)
  call CheckAllocation(allocErr, "norm (getSourceImage)")
  

  forall (i = 1 :nUseSource) norm(i) = 0.159154943 / (fitParams(2, i) * fitParams(3, i))
  

  newImage = pixelIgnoreFlag
  do i = 1, ny
     do j = 1, nx
        imageVal = image(j, i)
        do k = 1, nUseSource

           ! This bit can be speeded up by not using trig functions so much

           xdist = (j - fitParams(5, k))      
           ydist = (i - fitParams(6, k))
           radius(k) = sqrt(xdist ** 2 + ydist ** 2)
           angle = fitParams(4, k) - atan2(ydist, xdist) 
           xdist = radius(k) * cos(angle)
           ydist = radius(k) * sin(angle)
           flux(k) = fitParams(1, k) * exp(-0.5 * ((xdist / fitParams(2, k)) ** 2 + &
                (ydist / fitParams(3, k)) ** 2)) * norm(k)
           
        end do
        fluxTotal = 0.0
        do k = 1, nUseSource
           if(imageVal .gt. 0) then
              if(flux(k) / imageVal .lt. 0.05) then
                 flux(k) = -1
              else
                 fluxTotal = fluxTotal + flux(k)
              endif
           end if
        end do
        if(fluxTotal .gt. 0.0) then
           ! Assign fraction of value to each source image
           forall (k = 1 : nUseSource, flux(k) .ne. -1) newImage(j, i, k) = flux(k) * imageVal / fluxTotal
              
        end if
     end do
  end do

  deallocate(norm, stat = allocErr)
  call CheckDeallocation(allocErr, "norm (getSourceImage)")

end subroutine getSourceImage
!*************************************************************
! Subroutine getParametersfromMoments
! Purpose: Loops through all the sources to
! compute the source parameters- position,
! count  rates, etc
! Parameters:
! newImage - A 3-d array containing a 2-d image for each array.
! background - A 2-d background image.
! Returns: srcParams- a 1-d array of structures containing
! the parameters for each source. 
!*************************************************************
subroutine getParametersfromMoments(newImage, background, nx, ny, &
     nUseSource, srcParams, nSigma)

  integer, intent(in) :: nx, ny
  integer, intent(in) :: nUseSource
  real, intent(in) :: newImage(nx, ny, nUseSource)
  real, intent(in) :: background(nx, ny)
  type(sourceParams), pointer :: srcParams(:)
  real, intent(in) :: nSigma

  integer :: i
  real, allocatable :: aValue(:), bValue(:), theta(:)
  real, allocatable :: xPos(:), yPos(:), x2(:), y2(:), xy(:), signif(:)
  real, allocatable :: xPosErr(:), yPosErr(:), countrate(:), countrateErr(:)
  real, allocatable :: backGroundRate(:)   ! Added by CB to store the background counts
  integer(kind=int32), dimension(:), allocatable:: noSourcePixels ! Added by CB to store the number of pixels enclosed by the source
  integer :: allocErr

  real :: val1, val2
 
  allocate(xPos(nUseSource), stat=allocErr)
  call CheckAllocation(allocErr, "xPos (confusedSourceParams)")
  
  allocate(yPos(nUseSource), stat=allocErr)
  call CheckAllocation(allocErr, "yPos (confusedSourceParams)")
  
  allocate(xPosErr(nUseSource), stat=allocErr)
  call CheckAllocation(allocErr, "xPosErr (confusedSourceParams)")

  allocate(yPosErr(nUseSource), stat=allocErr)
  call CheckAllocation(allocErr, "yPosErr (confusedSourceParams)")
  
  allocate(x2(nUseSource), stat=allocErr)
  call CheckAllocation(allocErr, "pixel%next (x2 (confusedSourceParams)")

  allocate(y2(nUseSource), stat=allocErr)
  call CheckAllocation(allocErr, "y2 (confusedSourceParams)")
  
  allocate(xy(nUseSource), stat=allocErr)
  call CheckAllocation(allocErr, "xy (confusedSourceParams)")
 
  allocate(aValue(nUseSource), stat=allocErr)
  call CheckAllocation(allocErr, "aValue (confusedSourceParams)")
  
  allocate(bValue(nUseSource), stat=allocErr)
  call CheckAllocation(allocErr, "bValue (confusedSourceParams)")
  
  allocate(theta(nUseSource), stat=allocErr)
  call CheckAllocation(allocErr, "theta (confusedSourceParams)")
 
  allocate(countrate(nUseSource), stat=allocErr)
  call CheckAllocation(allocErr, "countrate (confusedSourceParams)")
  
  allocate(countrateErr(nUseSource),stat=allocErr)
  call CheckAllocation(allocErr, "countrateErr (confusedSourceParams)")
  
  allocate(backGroundRate(nUseSource),stat=allocErr)
  call CheckAllocation(allocErr, "backGroundRate (confusedSourceParams)")
  
  allocate(signif(nUseSource),stat=allocErr)
  call CheckAllocation(allocErr, "signif (confusedSourceParams)")
 
  allocate(noSourcePixels(1 : nUseSource),stat=allocErr)
  call CheckAllocation(allocErr, "noSourcePixels (confusedSourceParams)")
 

 
  !*****************************************************************
  ! Get the source position, countrates and background count rates
  ! for each source
  ! newImage is a 3-d array (nx, ny, nUseSource)
  !*****************************************************************
  call getMoments(newImage, background, nx, ny, nUseSource, xPos, Ypos, &
       xPosErr, yPosErr, x2, y2, xy, countrate, countrateErr, &
       signif, backGroundRate, noSourcePixels)

  !*****************************************************
  ! Now convert these into major/minor values and theta
  ! factor of 2 somewhere?
  !*****************************************************
  do i = 1, nUseSource
     if(countrate(i) .gt. 0 .and. signif(i) .ge. nSigma) then
        val1 = (x2(i) + y2(i)) / 2.0
        val2 = sqrt(((x2(i) - y2(i)) / 2.0) ** 2 + xy(i) ** 2)
        aValue(i) = sqrt(val1 + val2)

        if(val1 .gt. val2) then
           bValue(i) = sqrt(val1 - val2)
        
           !********************************* 
           ! Sort out angle
           !*********************************
           theta(i) = Angle(x2(i), y2(i), xy(i))
 
           !***************************************************
           ! Now assign fluxes to values in original levelImage 
           !**************************************************
           srcParams = sourceParams(xPos(i), yPos(i), xPosErr(i),           &
                                   yPosErr(i), theta(i) * radiansToDegrees, &
                                   aValue(i), bValue(i),                    &
                                   countrate(i), countrateErr(i),           &
                                   backGroundRate(i),                       &
                                   noSourcePixels(i), signif(i), 0)
 

           if(nUseSource .gt. 1) then
              srcParams(i)%Cflag = 1
           else
              srcParams(i)%Cflag = 0
           endif
        else
           write(errstr,'(''Warning: Problem with source parameterisation - ignoring source'')')
           call message(AppMsg, Verbose)   ! Verbose was 2
           srcParams(i)%countrate = -1
           srcParams(i)%backGroundRate = -1 !Source background rate
        endif
     else
        srcParams(i)%countrate = -1
        srcParams(i)%backGroundRate = -1   
     endif
  end do

  ! Clear up

  deallocate(xPos, stat = allocErr)
  call CheckDeallocation(allocErr, "xPos (getParametersfromMoments)")

  deallocate(yPos, stat = allocErr)
  call CheckDeallocation(allocErr, "yPos (getParametersfromMoments)")


  deallocate(xPosErr, stat = allocErr)
  call CheckDeallocation(allocErr, "xPosErr (getParametersfromMoments)")


  deallocate(yPosErr, stat = allocerr)
  call CheckDeallocation(allocErr, "yPosErr (getParametersfromMoments)")

  deallocate(x2, stat = allocErr)
  call CheckDeallocation(allocErr, "x2 (getParametersfromMoments)")

  deallocate(y2, stat = allocErr)
  call CheckDeallocation(allocErr, "y2 (getParametersfromMoments)")

  deallocate(xy, stat = allocErr)
  call CheckDeallocation(allocErr, "xy (getParametersfromMoments)")


  deallocate(countrate, stat = allocErr)
  call CheckDeallocation(allocErr, "countrate (getParametersfromMoments)")

  deallocate(countrateErr, stat = allocErr)
  call CheckDeallocation(allocErr, "countrateErr (getParametersfromMoments)")

  deallocate(signif, stat = allocErr)
  call CheckDeallocation(allocErr, "signif (getParametersfromMoments)")

  deallocate(aValue, stat = allocErr)
  call CheckDeallocation(allocErr, "aValue (getParametersfromMoments)")

  deallocate(bValue, stat = allocErr)
  call CheckDeallocation(allocErr, "bValue (getParametersfromMoments)")

  deallocate(theta, stat = allocErr)
  call CheckDeallocation(allocErr, "theta (getParametersfromMoments)")

  deallocate(backGroundRate, stat = allocErr)
  call CheckDeallocation(allocErr, "xPos (backGroundRate)")
  
  deallocate(noSourcePixels, stat = allocErr)
  call CheckDeallocation(allocErr, "noSourcePixels (backGroundRate)")


end subroutine getParametersfromMoments
!********************************************************
! subroutine getMoments
! Purpose: Loops through the number of sources found for each 
! source in the crude source list (this value will normally be 1
! and hence this routine will usually be called once for each 
! of the initially detected sources)
! Each source is parameterised using moments
! Returns the x and y positions, countrates 
! and background rates for each source
! Parameters:
! image - A 3-d array- one 2-d image for each source
!********************************************************
subroutine getMoments(image, background, nx, ny, nUseSource, xPos, yPos, xPosErr,&
     yPosErr, x2, y2, xy, countrate, countrateErr, &
     signif, backGroundRate, nPixels)
  
  integer, intent(in) :: nx, ny, nUseSource
  real, intent(in)    :: image(nx, ny, nUseSource)
  real, intent(in)    :: background(nx, ny)          
  real, intent(out)   :: xPos(nUseSource)
  real, intent(out)   :: yPos(nUseSource)
  real, intent(out)   :: xPosErr(nUseSource)
  real, intent(out)   :: yPosErr(nUseSource)
  real, intent(out)   :: x2(nUseSource)
  real, intent(out)   :: y2(nUseSource)
  real, intent(out)   :: xy(nUseSource)
  real, intent(out)   :: countrate(nUseSource)       ! Source count-rate
  real, intent(out)   :: countrateErr(nUseSource)    ! Source count-rate error
  real, intent(out)   :: signif(nUseSource)          ! Source significance
  real, intent(out)   :: backGroundRate(nUseSource)  ! Source background
  integer :: i, j, k
  real :: totalImage, totalBack, totalImage2, totalBackVal
  real :: imageVal
  integer(kind=int32), intent(out), dimension(1 : nUseSource) :: nPixels
  real :: xErrTot, yErrTot

!  write(*, *) "getMoments nUseSource=", nUseSource, " nx=", nx, " ny=",  ny


  do k = 1, nUseSource

     xpos(k) = 0.0
     ypos(k) = 0.0
     totalImage = 0.0
     totalImage2 = 0.0
     totalBack = 0.0
     nPixels(k) = 0

     do i = 1,  ny
        do j = 1, nx
           imageVal = image(j, i, k)

           if(imageVal .ne. pixelIgnoreFlag) then
              xpos(k) = xpos(k) + imageVal * j
              ypos(k) = ypos(k) + imageVal * i
              totalImage2 = totalImage2 + imageVal * imageVal
              totalImage = totalImage + imageVal
              nPixels(k) = nPixels(k) + 1
              totalback = totalback + background(j, i)
           end if
        end do
     end do

     if(totalImage .eq. 0) countrate(k) = -1
     

     totalBackVal = totalBack
     totalBack = totalBack / real(nPixels(k))
     xErrTot = 0.0
     yErrTot = 0.0
     xpos(k) = xpos(k) / totalImage
     ypos(k) = ypos(k) / totalImage

     xPosErr(k) = 0.0
     yPosErr(k) = 0.0
     x2(k) = 0.0
     y2(k) = 0.0
     xy(k) = 0.0


     do i = 1, ny
        do j = 1, nx
           imageVal = image(j, i, k)
           if(imageVal .ne. pixelIgnoreFlag) then

              !*******************************************************
              ! Assume that the error is dominated by the total counts
              !*******************************************************
              xErrTot = xErrTot + (imageVal + totalBack) * (j - xPos(k)) ** 2
              yErrTot = yErrTot + (imageVal + totalBack) * (i - yPos(k)) ** 2

              x2(k) = x2(k) + imageVal * j * j
              y2(k) = y2(k) + imageVal * i * i
              xy(k) = xy(k) + imageVal * i * j
           end if
        end do
     end do

     xPosErr(k) = sqrt(xErrTot / totalImage2)   ! Source pixel x-position error
     yPosErr(k) = sqrt(yErrTot / totalImage2)   ! Source pixel y-position error

     x2(k) = x2(k) / totalImage - xpos(k) * xpos(k)
     y2(k) = y2(k) / totalImage - ypos(k) * ypos(k)
     xy(k) = xy(k) / totalImage - xpos(k) * ypos(k)

     countrate(k) = totalImage                              ! Source total count-rate 
     countrateErr(k) = sqrt(totalImage + 2 * totalBackVal)  ! Error in countrate
     signif(k) = totalImage / sqrt(totalBackVal)            ! Source significance 

     backGroundRate(k) = totalBackVal     ! Background counts


  end do ! End of looping through the sources



end subroutine getMoments
!*************************************
! Routine to delete branch from Tree
!*************************************
recursive subroutine clearTree(Tree)

  type(treeStruct), pointer :: Tree

  ! Go down sources

  loop: do 
     if(associated(Tree%next)) then
        call clearTree(Tree%next)
     else
        exit loop
     end if
  end do loop

  ! Do down levels

  if(associated(Tree%downlevel)) call clearTree(Tree%downlevel)
  

  ! Remove tree from memory

  ! Remove pixelsmaxp etc.
  if(associated(Tree%levelImage)) then
     deallocate(Tree%levelImage)
     nullify(Tree%levelImage)
  end if

  deallocate(Tree)
  nullify(Tree)

end subroutine clearTree

end module omdetectSourceParams_mod

















